"""Revocation registry admin routes."""

import json
import logging
import os
import re
import shutil
from asyncio import shield

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields, validate, validates_schema
from marshmallow.exceptions import ValidationError

from ..admin.decorators.auth import tenant_authentication
from ..admin.request_context import AdminRequestContext
from ..connections.models.conn_record import ConnRecord
from ..core.event_bus import Event, EventBus
from ..core.profile import Profile
from ..indy.issuer import IndyIssuerError
from ..ledger.base import BaseLedger
from ..ledger.error import LedgerError
from ..ledger.multiple_ledger.base_manager import BaseMultipleLedgerManager
from ..messaging.credential_definitions.util import CRED_DEF_SENT_RECORD_TYPE
from ..messaging.models.base import BaseModelError
from ..messaging.models.openapi import OpenAPISchema
from ..messaging.responder import BaseResponder
from ..messaging.valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_CRED_REV_ID_EXAMPLE,
    INDY_CRED_REV_ID_VALIDATE,
    INDY_REV_REG_ID_EXAMPLE,
    INDY_REV_REG_ID_VALIDATE,
    INDY_REV_REG_SIZE_EXAMPLE,
    INDY_REV_REG_SIZE_VALIDATE,
    UUID4_EXAMPLE,
    UUID4_VALIDATE,
    WHOLE_NUM_EXAMPLE,
    WHOLE_NUM_VALIDATE,
)
from ..protocols.endorse_transaction.v1_0.manager import (
    TransactionManager,
    TransactionManagerError,
)
from ..protocols.endorse_transaction.v1_0.models.transaction_record import (
    TransactionRecordSchema,
)
from ..protocols.endorse_transaction.v1_0.util import (
    get_endorser_connection_id,
    is_author_role,
)
from ..storage.base import BaseStorage
from ..storage.error import StorageError, StorageNotFoundError
from ..utils.profiles import is_anoncreds_profile_raise_web_exception
from .error import RevocationError, RevocationNotSupportedError
from .indy import IndyRevocation
from .manager import RevocationManager, RevocationManagerError
from .models.issuer_cred_rev_record import IssuerCredRevRecord, IssuerCredRevRecordSchema
from .models.issuer_rev_reg_record import IssuerRevRegRecord, IssuerRevRegRecordSchema
from .util import (
    REVOCATION_ENTRY_EVENT,
    REVOCATION_EVENT_PREFIX,
    REVOCATION_REG_ENDORSED_EVENT,
    REVOCATION_REG_INIT_EVENT,
    notify_revocation_entry_event,
)

LOGGER = logging.getLogger(__name__)


class RevocationModuleResponseSchema(OpenAPISchema):
    """Response schema for Revocation Module."""


class RevRegCreateRequestSchema(OpenAPISchema):
    """Request schema for revocation registry creation request."""

    credential_definition_id = fields.Str(
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    max_cred_num = fields.Int(
        required=False,
        validate=INDY_REV_REG_SIZE_VALIDATE,
        metadata={
            "description": "Revocation registry size",
            "strict": True,
            "example": INDY_REV_REG_SIZE_EXAMPLE,
        },
    )


class RevRegResultSchema(OpenAPISchema):
    """Result schema for revocation registry creation request."""

    result = fields.Nested(IssuerRevRegRecordSchema())


class TxnOrRevRegResultSchema(OpenAPISchema):
    """Result schema for credential definition send request."""

    sent = fields.Nested(
        RevRegResultSchema(), required=False, metadata={"definition": "Content sent"}
    )
    txn = fields.Nested(
        TransactionRecordSchema(),
        required=False,
        metadata={"description": "Revocation registry definition transaction to endorse"},
    )


class CredRevRecordQueryStringSchema(OpenAPISchema):
    """Parameters and validators for credential revocation record request."""

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields - must have (rr-id and cr-id) xor cx-id."""

        rev_reg_id = data.get("rev_reg_id")
        cred_rev_id = data.get("cred_rev_id")
        cred_ex_id = data.get("cred_ex_id")

        if not (
            (rev_reg_id and cred_rev_id and not cred_ex_id)
            or (cred_ex_id and not rev_reg_id and not cred_rev_id)
        ):
            raise ValidationError(
                "Request must have either rev_reg_id and cred_rev_id or cred_ex_id"
            )

    rev_reg_id = fields.Str(
        required=False,
        validate=INDY_REV_REG_ID_VALIDATE,
        metadata={
            "description": "Revocation registry identifier",
            "example": INDY_REV_REG_ID_EXAMPLE,
        },
    )
    cred_rev_id = fields.Str(
        required=False,
        validate=INDY_CRED_REV_ID_VALIDATE,
        metadata={
            "description": "Credential revocation identifier",
            "example": INDY_CRED_REV_ID_EXAMPLE,
        },
    )
    cred_ex_id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Credential exchange identifier",
            "example": UUID4_EXAMPLE,
        },
    )


class RevRegId(OpenAPISchema):
    """Parameters and validators for delete tails file request."""

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields - must have either rr-id or cr-id."""

        rev_reg_id = data.get("rev_reg_id")
        cred_def_id = data.get("cred_def_id")

        if not (rev_reg_id or cred_def_id):
            raise ValidationError("Request must have either rev_reg_id or cred_def_id")

    rev_reg_id = fields.Str(
        required=False,
        validate=INDY_REV_REG_ID_VALIDATE,
        metadata={
            "description": "Revocation registry identifier",
            "example": INDY_REV_REG_ID_EXAMPLE,
        },
    )
    cred_def_id = fields.Str(
        required=False,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )


class RevokeRequestSchema(CredRevRecordQueryStringSchema):
    """Parameters and validators for revocation request."""

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate fields - connection_id and thread_id must be present if notify."""
        super().validate_fields(data, **kwargs)

        notify = data.get("notify")
        connection_id = data.get("connection_id")
        notify_version = data.get("notify_version", "v1_0")

        if notify and not connection_id:
            raise ValidationError("Request must specify connection_id if notify is true")
        if notify and not notify_version:
            raise ValidationError("Request must specify notify_version if notify is true")

    publish = fields.Boolean(
        required=False,
        metadata={
            "description": (
                "(True) publish revocation to ledger immediately, or (default, False)"
                " mark it pending"
            )
        },
    )
    notify = fields.Boolean(
        required=False,
        metadata={"description": "Send a notification to the credential recipient"},
    )
    notify_version = fields.String(
        validate=validate.OneOf(["v1_0", "v2_0"]),
        required=False,
        metadata={
            "description": (
                "Specify which version of the revocation notification should be sent"
            )
        },
    )
    connection_id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": (
                "Connection ID to which the revocation notification will be sent;"
                " required if notify is true"
            ),
            "example": UUID4_EXAMPLE,
        },
    )
    thread_id = fields.Str(
        required=False,
        metadata={
            "description": (
                "Thread ID of the credential exchange message thread resulting in the"
                " credential now being revoked; required if notify is true"
            )
        },
    )
    comment = fields.Str(
        required=False,
        metadata={
            "description": "Optional comment to include in revocation notification"
        },
    )


class PublishRevocationsSchema(OpenAPISchema):
    """Request and result schema for revocation publication API call."""

    rrid2crid = fields.Dict(
        required=False,
        keys=fields.Str(metadata={"example": INDY_REV_REG_ID_EXAMPLE}),
        values=fields.List(
            fields.Str(
                validate=INDY_CRED_REV_ID_VALIDATE,
                metadata={
                    "description": "Credential revocation identifier",
                    "example": INDY_CRED_REV_ID_EXAMPLE,
                },
            )
        ),
        metadata={"description": "Credential revocation ids by revocation registry id"},
    )


class TxnOrPublishRevocationsResultSchema(PublishRevocationsSchema):
    """Result schema for credential definition send request."""

    txn = fields.List(
        fields.Nested(
            TransactionRecordSchema(),
            required=False,
            metadata={
                "description": "Revocation registry revocations transaction to endorse"
            },
        )
    )


class ClearPendingRevocationsRequestSchema(OpenAPISchema):
    """Request schema for clear pending revocations API call."""

    purge = fields.Dict(
        required=False,
        keys=fields.Str(metadata={"example": INDY_REV_REG_ID_EXAMPLE}),
        values=fields.List(
            fields.Str(
                validate=INDY_CRED_REV_ID_VALIDATE,
                metadata={
                    "description": "Credential revocation identifier",
                    "example": INDY_CRED_REV_ID_EXAMPLE,
                },
            )
        ),
        metadata={
            "description": (
                "Credential revocation ids by revocation registry id: omit for all,"
                " specify null or empty list for all pending per revocation registry"
            )
        },
    )


class CredRevRecordResultSchema(OpenAPISchema):
    """Result schema for credential revocation record request."""

    result = fields.Nested(IssuerCredRevRecordSchema())


class CredRevRecordDetailsResultSchema(OpenAPISchema):
    """Result schema for credential revocation record request."""

    results = fields.List(fields.Nested(IssuerCredRevRecordSchema()))


class CredRevIndyRecordsResultSchema(OpenAPISchema):
    """Result schema for revoc reg delta."""

    rev_reg_delta = fields.Dict(
        metadata={"description": "Indy revocation registry delta"}
    )


class RevRegIssuedResultSchema(OpenAPISchema):
    """Result schema for revocation registry credentials issued request."""

    result = fields.Int(
        validate=WHOLE_NUM_VALIDATE,
        metadata={
            "description": "Number of credentials issued against revocation registry",
            "strict": True,
            "example": WHOLE_NUM_EXAMPLE,
        },
    )


class RevRegUpdateRequestMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking rev reg id."""

    apply_ledger_update = fields.Bool(
        required=True,
        metadata={"description": "Apply updated accumulator transaction to ledger"},
    )


class RevRegWalletUpdatedResultSchema(OpenAPISchema):
    """Number of wallet revocation entries status updated."""

    rev_reg_delta = fields.Dict(
        metadata={"description": "Indy revocation registry delta"}
    )
    accum_calculated = fields.Dict(
        metadata={"description": "Calculated accumulator for phantom revocations"}
    )
    accum_fixed = fields.Dict(
        metadata={"description": "Applied ledger transaction to fix revocations"}
    )


class RevRegsCreatedSchema(OpenAPISchema):
    """Result schema for request for revocation registries created."""

    rev_reg_ids = fields.List(
        fields.Str(
            validate=INDY_REV_REG_ID_VALIDATE,
            metadata={
                "description": "Revocation registry identifiers",
                "example": INDY_REV_REG_ID_EXAMPLE,
            },
        )
    )


class RevRegUpdateTailsFileUriSchema(OpenAPISchema):
    """Request schema for updating tails file URI."""

    tails_public_uri = fields.Url(
        required=True,
        metadata={
            "description": "Public URI to the tails file",
            "example": (
                "http://192.168.56.133:6543/revocation/registry/"
                f"{INDY_REV_REG_ID_EXAMPLE}/tails-file"
            ),
        },
    )


class RevRegsCreatedQueryStringSchema(OpenAPISchema):
    """Query string parameters and validators for rev regs created request."""

    cred_def_id = fields.Str(
        required=False,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    state = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                getattr(IssuerRevRegRecord, m)
                for m in vars(IssuerRevRegRecord)
                if m.startswith("STATE_")
            ]
        ),
        metadata={"description": "Revocation registry state"},
    )


class SetRevRegStateQueryStringSchema(OpenAPISchema):
    """Query string parameters and validators for request to set rev reg state."""

    state = fields.Str(
        required=True,
        validate=validate.OneOf(
            [
                getattr(IssuerRevRegRecord, m)
                for m in vars(IssuerRevRegRecord)
                if m.startswith("STATE_") and m != "STATE_DECOMMISSIONED"
            ]
        ),
        metadata={"description": "Revocation registry state to set"},
    )


class RevRegIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking rev reg id."""

    rev_reg_id = fields.Str(
        required=True,
        validate=INDY_REV_REG_ID_VALIDATE,
        metadata={
            "description": "Revocation Registry identifier",
            "example": INDY_REV_REG_ID_EXAMPLE,
        },
    )


class RevocationCredDefIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking cred def id."""

    cred_def_id = fields.Str(
        required=True,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )


class CreateRevRegTxnForEndorserOptionSchema(OpenAPISchema):
    """Class for user to input whether to create a transaction for endorser or not."""

    create_transaction_for_endorser = fields.Boolean(
        required=False,
        metadata={"description": "Create Transaction For Endorser's signature"},
    )


class RevRegConnIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection id."""

    conn_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


@docs(
    tags=["revocation"],
    summary="Revoke an issued credential",
)
@request_schema(RevokeRequestSchema())
@querystring_schema(CreateRevRegTxnForEndorserOptionSchema())
@querystring_schema(RevRegConnIdMatchInfoSchema())
@response_schema(RevocationModuleResponseSchema(), description="")
@tenant_authentication
async def revoke(request: web.BaseRequest):
    """Request handler for storing a credential revocation.

    Args:
        request: aiohttp request object

    Returns:
        The credential revocation details.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    body = await request.json()
    cred_ex_id = body.get("cred_ex_id")
    body["notify"] = body.get("notify", context.settings.get("revocation.notify"))
    notify = body.get("notify", False)
    connection_id = body.get("connection_id")
    body["notify_version"] = body.get("notify_version", "v1_0")
    notify_version = body["notify_version"]
    create_transaction_for_endorser = json.loads(
        request.query.get("create_transaction_for_endorser", "false")
    )
    endorser_conn_id = request.query.get("conn_id")
    rev_manager = RevocationManager(profile)
    outbound_handler = request["outbound_message_router"]
    write_ledger = not create_transaction_for_endorser

    if is_author_role(profile):
        write_ledger = False
        create_transaction_for_endorser = True
        if not endorser_conn_id:
            endorser_conn_id = await get_endorser_connection_id(profile)
            if not endorser_conn_id:
                raise web.HTTPBadRequest(reason="No endorser connection found")
    if notify and not connection_id:
        raise web.HTTPBadRequest(reason="connection_id must be set when notify is true")
    if notify and not notify_version:
        raise web.HTTPBadRequest(
            reason="Request must specify notify_version if notify is true"
        )
    try:
        if cred_ex_id:
            rev_entry_resp = await rev_manager.revoke_credential_by_cred_ex_id(
                cred_ex_id=cred_ex_id,
                publish=body.get("publish", False),
                notify=notify,
                notify_version=notify_version,
                thread_id=body.get("thread_id"),
                connection_id=connection_id,
                endorser_conn_id=endorser_conn_id,
                comment=body.get("comment"),
                write_ledger=write_ledger,
            )
        else:
            rev_entry_resp = await rev_manager.revoke_credential(
                rev_reg_id=body.get("rev_reg_id"),
                cred_rev_id=body.get("cred_rev_id"),
                publish=body.get("publish", False),
                notify=notify,
                notify_version=notify_version,
                thread_id=body.get("thread_id"),
                connection_id=connection_id,
                endorser_conn_id=endorser_conn_id,
                comment=body.get("comment"),
                write_ledger=write_ledger,
            )
    except (
        RevocationManagerError,
        RevocationError,
        StorageError,
        IndyIssuerError,
        LedgerError,
    ) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    if create_transaction_for_endorser and rev_entry_resp:
        transaction_mgr = TransactionManager(profile)
        try:
            transaction = await transaction_mgr.create_record(
                messages_attach=rev_entry_resp["result"], connection_id=endorser_conn_id
            )
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        # if auto-request, send the request to the endorser
        if context.settings.get_value("endorser.auto_request"):
            try:
                (
                    transaction,
                    transaction_request,
                ) = await transaction_mgr.create_request(
                    transaction=transaction,
                )
            except (StorageError, TransactionManagerError) as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

            await outbound_handler(transaction_request, connection_id=endorser_conn_id)

        return web.json_response({"txn": transaction.serialize()})
    return web.json_response({})


@docs(tags=["revocation"], summary="Publish pending revocations to ledger")
@request_schema(PublishRevocationsSchema())
@querystring_schema(CreateRevRegTxnForEndorserOptionSchema())
@querystring_schema(RevRegConnIdMatchInfoSchema())
@response_schema(TxnOrPublishRevocationsResultSchema(), 200, description="")
@tenant_authentication
async def publish_revocations(request: web.BaseRequest):
    """Request handler for publishing pending revocations to the ledger.

    Args:
        request: aiohttp request object

    Returns:
        Credential revocation ids published as revoked by revocation registry id.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    body = await request.json()
    rrid2crid = body.get("rrid2crid")
    create_transaction_for_endorser = json.loads(
        request.query.get("create_transaction_for_endorser", "false")
    )
    write_ledger = not create_transaction_for_endorser
    endorser_conn_id = request.query.get("conn_id")
    rev_manager = RevocationManager(profile)
    outbound_handler = request["outbound_message_router"]

    if is_author_role(profile):
        write_ledger = False
        create_transaction_for_endorser = True
        endorser_conn_id = await get_endorser_connection_id(profile)
        if not endorser_conn_id:
            raise web.HTTPBadRequest(reason="No endorser connection found")
    try:
        rev_reg_responses, result = await rev_manager.publish_pending_revocations(
            rrid2crid=rrid2crid,
            write_ledger=write_ledger,
            connection_id=endorser_conn_id,
        )
    except (RevocationError, StorageError, IndyIssuerError, LedgerError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    if create_transaction_for_endorser:
        list_of_txns = await _process_publish_response_for_endorsement(
            profile, rev_reg_responses, outbound_handler, endorser_conn_id
        )
        return web.json_response({"txn": list_of_txns})

    return web.json_response({"rrid2crid": result})


async def _process_publish_response_for_endorsement(
    profile: Profile,
    responses: dict,
    outbound_handler: BaseResponder.send_outbound,
    endorser_conn_id: str,
):
    txn_responses = []
    for response in responses:
        transaction_mgr = TransactionManager(profile)
        try:
            transaction = await transaction_mgr.create_record(
                messages_attach=response["result"], connection_id=endorser_conn_id
            )
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        # if auto-request, send the request to the endorser
        if profile.context.settings.get_value("endorser.auto_request"):
            try:
                (
                    transaction,
                    transaction_request,
                ) = await transaction_mgr.create_request(
                    transaction=transaction,
                )
            except (StorageError, TransactionManagerError) as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

            await outbound_handler(transaction_request, connection_id=endorser_conn_id)

        txn_responses.append(transaction.serialize())

    return txn_responses


@docs(tags=["revocation"], summary="Clear pending revocations")
@request_schema(ClearPendingRevocationsRequestSchema())
@response_schema(PublishRevocationsSchema(), 200, description="")
@tenant_authentication
async def clear_pending_revocations(request: web.BaseRequest):
    """Request handler for clearing pending revocations.

    Args:
        request: aiohttp request object

    Returns:
        Credential revocation ids still pending revocation by revocation registry id.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    body = await request.json()
    purge = body.get("purge")

    rev_manager = RevocationManager(profile)

    try:
        results = await rev_manager.clear_pending_revocations(purge)
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    return web.json_response({"rrid2crid": results})


@docs(tags=["revocation"], summary="Rotate revocation registry")
@match_info_schema(RevocationCredDefIdMatchInfoSchema())
@response_schema(RevRegsCreatedSchema(), 200, description="")
@tenant_authentication
async def rotate_rev_reg(request: web.BaseRequest):
    """Request handler to rotate the active revocation registries for cred. def.

    Args:
        request: aiohttp request object

    Returns:
        list or revocation registry ids that were rotated out

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    cred_def_id = request.match_info["cred_def_id"]

    try:
        revoc = IndyRevocation(profile)
        recs = await revoc.decommission_registry(cred_def_id)
        del revoc
    except RevocationNotSupportedError as e:
        raise web.HTTPBadRequest(reason=e.message) from e

    return web.json_response(
        {"rev_reg_ids": [rec.revoc_reg_id for rec in recs if rec.revoc_reg_id]}
    )


@docs(tags=["revocation"], summary="Creates a new revocation registry")
@request_schema(RevRegCreateRequestSchema())
@response_schema(RevRegResultSchema(), 200, description="")
@tenant_authentication
async def create_rev_reg(request: web.BaseRequest):
    """Request handler to create a new revocation registry.

    Args:
        request: aiohttp request object

    Returns:
        The issuer revocation registry record

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    body = await request.json()

    credential_definition_id = body.get("credential_definition_id")
    max_cred_num = body.get("max_cred_num")

    # check we published this cred def
    async with profile.session() as session:
        storage = session.inject(BaseStorage)

        found = await storage.find_all_records(
            type_filter=CRED_DEF_SENT_RECORD_TYPE,
            tag_query={"cred_def_id": credential_definition_id},
        )
    if not found:
        raise web.HTTPNotFound(
            reason=f"Not issuer of credential definition id {credential_definition_id}"
        )

    try:
        revoc = IndyRevocation(profile)
        issuer_rev_reg_rec = await revoc.init_issuer_registry(
            credential_definition_id,
            max_cred_num=max_cred_num,
            notify=False,
        )
    except RevocationNotSupportedError as e:
        raise web.HTTPBadRequest(reason=e.message) from e
    await shield(issuer_rev_reg_rec.generate_registry(profile))

    return web.json_response({"result": issuer_rev_reg_rec.serialize()})


@docs(
    tags=["revocation"],
    summary="Search for matching revocation registries that current agent created",
)
@querystring_schema(RevRegsCreatedQueryStringSchema())
@response_schema(RevRegsCreatedSchema(), 200, description="")
@tenant_authentication
async def rev_regs_created(request: web.BaseRequest):
    """Request handler to get revocation registries that current agent created.

    Args:
        request: aiohttp request object

    Returns:
        List of identifiers of matching revocation registries.

    """
    context: AdminRequestContext = request["context"]

    is_anoncreds_profile_raise_web_exception(context.profile)

    search_tags = list(vars(RevRegsCreatedQueryStringSchema)["_declared_fields"])
    tag_filter = {tag: request.query[tag] for tag in search_tags if tag in request.query}
    async with context.profile.session() as session:
        found = await IssuerRevRegRecord.query(
            session,
            tag_filter,
            post_filter_negative={"state": IssuerRevRegRecord.STATE_INIT},
        )

    return web.json_response(
        {"rev_reg_ids": [record.revoc_reg_id for record in found if record.revoc_reg_id]}
    )


@docs(
    tags=["revocation"],
    summary="Get revocation registry by revocation registry id",
)
@match_info_schema(RevRegIdMatchInfoSchema())
@response_schema(RevRegResultSchema(), 200, description="")
@tenant_authentication
async def get_rev_reg(request: web.BaseRequest):
    """Request handler to get a revocation registry by rev reg id.

    Args:
        request: aiohttp request object

    Returns:
        The revocation registry

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]

    try:
        revoc = IndyRevocation(profile)
        rev_reg = await revoc.get_issuer_rev_reg_record(rev_reg_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    return web.json_response({"result": rev_reg.serialize()})


@docs(
    tags=["revocation"],
    summary="Get number of credentials issued against revocation registry",
)
@match_info_schema(RevRegIdMatchInfoSchema())
@response_schema(RevRegIssuedResultSchema(), 200, description="")
@tenant_authentication
async def get_rev_reg_issued_count(request: web.BaseRequest):
    """Request handler to get number of credentials issued against revocation registry.

    Args:
        request: aiohttp request object

    Returns:
        Number of credentials issued against revocation registry

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]

    async with profile.session() as session:
        try:
            await IssuerRevRegRecord.retrieve_by_revoc_reg_id(session, rev_reg_id)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        count = len(
            await IssuerCredRevRecord.query_by_ids(session, rev_reg_id=rev_reg_id)
        )

    return web.json_response({"result": count})


@docs(
    tags=["revocation"],
    summary="Get details of credentials issued against revocation registry",
)
@match_info_schema(RevRegIdMatchInfoSchema())
@response_schema(CredRevRecordDetailsResultSchema(), 200, description="")
@tenant_authentication
async def get_rev_reg_issued(request: web.BaseRequest):
    """Request handler to get credentials issued against revocation registry.

    Args:
        request: aiohttp request object

    Returns:
        Number of credentials issued against revocation registry

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]

    recs = []
    async with profile.session() as session:
        try:
            await IssuerRevRegRecord.retrieve_by_revoc_reg_id(session, rev_reg_id)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        recs = await IssuerCredRevRecord.query_by_ids(session, rev_reg_id=rev_reg_id)
    results = []
    for rec in recs:
        results.append(rec.serialize())

    return web.json_response(results)


@docs(
    tags=["revocation"],
    summary="Get details of revoked credentials from ledger",
)
@match_info_schema(RevRegIdMatchInfoSchema())
@response_schema(CredRevIndyRecordsResultSchema(), 200, description="")
@tenant_authentication
async def get_rev_reg_indy_recs(request: web.BaseRequest):
    """Request handler to get details of revoked credentials from ledger.

    Args:
        request: aiohttp request object

    Returns:
        Details of revoked credentials from ledger

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]

    revoc = IndyRevocation(profile)
    rev_reg_delta = await revoc.get_issuer_rev_reg_delta(rev_reg_id)

    return web.json_response(
        {
            "rev_reg_delta": rev_reg_delta,
        }
    )


@docs(
    tags=["revocation"],
    summary="Fix revocation state in wallet and return number of updated entries",
)
@match_info_schema(RevRegIdMatchInfoSchema())
@querystring_schema(RevRegUpdateRequestMatchInfoSchema())
@response_schema(RevRegWalletUpdatedResultSchema(), 200, description="")
@tenant_authentication
async def update_rev_reg_revoked_state(request: web.BaseRequest):
    """Request handler to fix ledger entry of credentials revoked against registry.

    Args:
        request: aiohttp request object

    Returns:
        Number of credentials posted to ledger

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    apply_ledger_update = json.loads(request.query.get("apply_ledger_update", "false"))
    LOGGER.debug(
        "Update revocation state request for rev_reg_id = %s, apply_ledger_update = %s",
        rev_reg_id,
        apply_ledger_update,
    )

    rev_reg_record = None
    genesis_transactions = None
    async with profile.session() as session:
        try:
            rev_reg_record = await IssuerRevRegRecord.retrieve_by_revoc_reg_id(
                session, rev_reg_id
            )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err

        genesis_transactions = context.settings.get("ledger.genesis_transactions")
        if not genesis_transactions:
            ledger_manager = context.injector.inject(BaseMultipleLedgerManager)
            write_ledger = context.injector.inject(BaseLedger)
            available_write_ledgers = await ledger_manager.get_write_ledgers()
            pool = write_ledger.pool
            genesis_transactions = pool.genesis_txns
            LOGGER.debug("available write_ledgers = %s", available_write_ledgers)
            LOGGER.debug("write_ledger = %s", write_ledger)
            LOGGER.debug("write_ledger pool = %s", pool)

        if not genesis_transactions:
            raise web.HTTPInternalServerError(
                reason="no genesis_transactions for writable ledger"
            )

        if apply_ledger_update:
            ledger = session.inject_or(BaseLedger)
            if not ledger:
                raise web.HTTPInternalServerError(
                    reason="No ledger available when requesting ledger update"
                )

    rev_manager = RevocationManager(profile)
    try:
        (
            rev_reg_delta,
            recovery_txn,
            applied_txn,
        ) = await rev_manager.update_rev_reg_revoked_state(
            apply_ledger_update, rev_reg_record, genesis_transactions
        )
    except (
        RevocationManagerError,
        RevocationError,
        StorageError,
        IndyIssuerError,
        LedgerError,
    ) as err:
        raise web.HTTPBadRequest(reason=err.roll_up)
    except Exception as err:
        raise web.HTTPBadRequest(reason=str(err))

    return web.json_response(
        {
            "rev_reg_delta": rev_reg_delta,
            "accum_calculated": recovery_txn,
            "accum_fixed": applied_txn,
        }
    )


@docs(
    tags=["revocation"],
    summary="Get credential revocation status",
)
@querystring_schema(CredRevRecordQueryStringSchema())
@response_schema(CredRevRecordResultSchema(), 200, description="")
@tenant_authentication
async def get_cred_rev_record(request: web.BaseRequest):
    """Request handler to get credential revocation record.

    Args:
        request: aiohttp request object

    Returns:
        The issuer credential revocation record

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.query.get("rev_reg_id")
    cred_rev_id = request.query.get("cred_rev_id")  # numeric string
    cred_ex_id = request.query.get("cred_ex_id")

    try:
        async with profile.session() as session:
            if rev_reg_id and cred_rev_id:
                rec = await IssuerCredRevRecord.retrieve_by_ids(
                    session, rev_reg_id, cred_rev_id
                )
            else:
                rec = await IssuerCredRevRecord.retrieve_by_cred_ex_id(
                    session, cred_ex_id
                )
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    return web.json_response({"result": rec.serialize()})


@docs(
    tags=["revocation"],
    summary="Get current active revocation registry by credential definition id",
)
@match_info_schema(RevocationCredDefIdMatchInfoSchema())
@response_schema(RevRegResultSchema(), 200, description="")
@tenant_authentication
async def get_active_rev_reg(request: web.BaseRequest):
    """Request handler to get current active revocation registry by cred def id.

    Args:
        request: aiohttp request object

    Returns:
        The revocation registry identifier

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    cred_def_id = request.match_info["cred_def_id"]

    try:
        revoc = IndyRevocation(profile)
        rev_reg = await revoc.get_active_issuer_rev_reg_record(cred_def_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    return web.json_response({"result": rev_reg.serialize()})


@docs(
    tags=["revocation"],
    summary="Download tails file",
    produces=["application/octet-stream"],
)
@match_info_schema(RevRegIdMatchInfoSchema())
@response_schema(RevocationModuleResponseSchema, description="tails file")
@tenant_authentication
async def get_tails_file(request: web.BaseRequest) -> web.FileResponse:
    """Request handler to download tails file for revocation registry.

    Args:
        request: aiohttp request object

    Returns:
        The tails file in FileResponse

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]

    try:
        revoc = IndyRevocation(profile)
        rev_reg = await revoc.get_issuer_rev_reg_record(rev_reg_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    return web.FileResponse(path=rev_reg.tails_local_path, status=200)


@docs(
    tags=["revocation"],
    summary="Upload local tails file to server",
)
@match_info_schema(RevRegIdMatchInfoSchema())
@response_schema(RevocationModuleResponseSchema(), description="")
@tenant_authentication
async def upload_tails_file(request: web.BaseRequest):
    """Request handler to upload local tails file for revocation registry.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    try:
        revoc = IndyRevocation(profile)
        rev_reg = await revoc.get_issuer_rev_reg_record(rev_reg_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    if not rev_reg.has_local_tails_file:
        raise web.HTTPNotFound(reason=f"No local tails file for rev reg {rev_reg_id}")

    try:
        await rev_reg.upload_tails_file(profile)
    except RevocationError as e:
        raise web.HTTPInternalServerError(reason=str(e))

    return web.json_response({})


@docs(
    tags=["revocation"],
    summary="Send revocation registry definition to ledger",
)
@match_info_schema(RevRegIdMatchInfoSchema())
@querystring_schema(CreateRevRegTxnForEndorserOptionSchema())
@querystring_schema(RevRegConnIdMatchInfoSchema())
@response_schema(TxnOrRevRegResultSchema(), 200, description="")
@tenant_authentication
async def send_rev_reg_def(request: web.BaseRequest):
    """Request handler to send revocation registry definition by rev reg id to ledger.

    Args:
        request: aiohttp request object

    Returns:
        The issuer revocation registry record

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    outbound_handler = request["outbound_message_router"]
    rev_reg_id = request.match_info["rev_reg_id"]
    create_transaction_for_endorser = json.loads(
        request.query.get("create_transaction_for_endorser", "false")
    )
    write_ledger = not create_transaction_for_endorser
    endorser_did = None
    connection_id = request.query.get("conn_id")

    # check if we need to endorse
    if is_author_role(profile):
        # authors cannot write to the ledger
        write_ledger = False
        create_transaction_for_endorser = True
        if not connection_id:
            # author has not provided a connection id, so determine which to use
            connection_id = await get_endorser_connection_id(profile)
            if not connection_id:
                raise web.HTTPBadRequest(reason="No endorser connection found")

    if not write_ledger:
        try:
            async with profile.session() as session:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, connection_id
                )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except BaseModelError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        async with profile.session() as session:
            endorser_info = await connection_record.metadata_get(session, "endorser_info")
        if not endorser_info:
            raise web.HTTPForbidden(
                reason=(
                    "Endorser Info is not set up in "
                    "connection metadata for this connection record"
                )
            )
        if "endorser_did" not in endorser_info.keys():
            raise web.HTTPForbidden(
                reason=(
                    ' "endorser_did" is not set in "endorser_info"'
                    " in connection metadata for this connection record"
                )
            )
        endorser_did = endorser_info["endorser_did"]

    try:
        revoc = IndyRevocation(profile)
        rev_reg = await revoc.get_issuer_rev_reg_record(rev_reg_id)

        rev_reg_resp = await rev_reg.send_def(
            profile,
            write_ledger=write_ledger,
            endorser_did=endorser_did,
        )
        LOGGER.debug("published rev reg definition: %s", rev_reg_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except RevocationError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    if not create_transaction_for_endorser:
        return web.json_response({"sent": rev_reg.serialize()})

    else:
        transaction_mgr = TransactionManager(profile)
        try:
            transaction = await transaction_mgr.create_record(
                messages_attach=rev_reg_resp["result"], connection_id=connection_id
            )
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        # if auto-request, send the request to the endorser
        if context.settings.get_value("endorser.auto_request"):
            try:
                (
                    transaction,
                    transaction_request,
                ) = await transaction_mgr.create_request(
                    transaction=transaction,
                    # TODO see if we need to parametrize these params
                    # expires_time=expires_time,
                )
            except (StorageError, TransactionManagerError) as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

            await outbound_handler(transaction_request, connection_id=connection_id)

        return web.json_response({"txn": transaction.serialize()})


@docs(
    tags=["revocation"],
    summary="Send revocation registry entry to ledger",
)
@match_info_schema(RevRegIdMatchInfoSchema())
@querystring_schema(CreateRevRegTxnForEndorserOptionSchema())
@querystring_schema(RevRegConnIdMatchInfoSchema())
@response_schema(RevRegResultSchema(), 200, description="")
@tenant_authentication
async def send_rev_reg_entry(request: web.BaseRequest):
    """Request handler to send rev reg entry by registry id to ledger.

    Args:
        request: aiohttp request object

    Returns:
        The revocation registry record

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    outbound_handler = request["outbound_message_router"]
    create_transaction_for_endorser = json.loads(
        request.query.get("create_transaction_for_endorser", "false")
    )
    write_ledger = not create_transaction_for_endorser
    endorser_did = None
    connection_id = request.query.get("conn_id")
    rev_reg_id = request.match_info["rev_reg_id"]

    # check if we need to endorse
    if is_author_role(profile):
        # authors cannot write to the ledger
        write_ledger = False
        create_transaction_for_endorser = True
        if not connection_id:
            # author has not provided a connection id, so determine which to use
            connection_id = await get_endorser_connection_id(profile)
            if not connection_id:
                raise web.HTTPBadRequest(reason="No endorser connection found")

    if not write_ledger:
        async with profile.session() as session:
            try:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, connection_id
                )
            except StorageNotFoundError as err:
                raise web.HTTPNotFound(reason=err.roll_up) from err
            except BaseModelError as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

            endorser_info = await connection_record.metadata_get(session, "endorser_info")
        if not endorser_info:
            raise web.HTTPForbidden(
                reason=(
                    "Endorser Info is not set up in "
                    "connection metadata for this connection record"
                )
            )
        if "endorser_did" not in endorser_info.keys():
            raise web.HTTPForbidden(
                reason=(
                    ' "endorser_did" is not set in "endorser_info"'
                    " in connection metadata for this connection record"
                )
            )
        endorser_did = endorser_info["endorser_did"]

    try:
        revoc = IndyRevocation(profile)
        rev_reg = await revoc.get_issuer_rev_reg_record(rev_reg_id)
        rev_entry_resp = await rev_reg.send_entry(
            profile,
            write_ledger=write_ledger,
            endorser_did=endorser_did,
        )
        LOGGER.debug("published registry entry: %s", rev_reg_id)

    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except RevocationError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    if not create_transaction_for_endorser:
        return web.json_response({"result": rev_reg.serialize()})

    else:
        transaction_mgr = TransactionManager(profile)
        try:
            transaction = await transaction_mgr.create_record(
                messages_attach=rev_entry_resp["result"],
                connection_id=connection_id,
            )
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        # if auto-request, send the request to the endorser
        if context.settings.get_value("endorser.auto_request"):
            try:
                (
                    transaction,
                    transaction_request,
                ) = await transaction_mgr.create_request(
                    transaction=transaction,
                    # TODO see if we need to parametrize these params
                    # expires_time=expires_time,
                )
            except (StorageError, TransactionManagerError) as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

            await outbound_handler(transaction_request, connection_id=connection_id)

        return web.json_response({"txn": transaction.serialize()})


@docs(
    tags=["revocation"],
    summary="Update revocation registry with new public URI to its tails file",
)
@match_info_schema(RevRegIdMatchInfoSchema())
@request_schema(RevRegUpdateTailsFileUriSchema())
@response_schema(RevRegResultSchema(), 200, description="")
@tenant_authentication
async def update_rev_reg(request: web.BaseRequest):
    """Request handler to update a rev reg's public tails URI by registry id.

    Args:
        request: aiohttp request object

    Returns:
        The revocation registry record

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    body = await request.json()
    tails_public_uri = body.get("tails_public_uri")

    rev_reg_id = request.match_info["rev_reg_id"]

    try:
        revoc = IndyRevocation(profile)
        rev_reg = await revoc.get_issuer_rev_reg_record(rev_reg_id)
        await rev_reg.set_tails_file_public_uri(profile, tails_public_uri)

    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except RevocationError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"result": rev_reg.serialize()})


@docs(tags=["revocation"], summary="Set revocation registry state manually")
@match_info_schema(RevRegIdMatchInfoSchema())
@querystring_schema(SetRevRegStateQueryStringSchema())
@response_schema(RevRegResultSchema(), 200, description="")
@tenant_authentication
async def set_rev_reg_state(request: web.BaseRequest):
    """Request handler to set a revocation registry state manually.

    Args:
        request: aiohttp request object

    Returns:
        The revocation registry record, updated

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    state = request.query.get("state")

    try:
        revoc = IndyRevocation(profile)
        rev_reg = await revoc.get_issuer_rev_reg_record(rev_reg_id)
        async with profile.session() as session:
            await rev_reg.set_state(session, state)

        LOGGER.debug("set registry %s state: %s", rev_reg_id, state)

    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    return web.json_response({"result": rev_reg.serialize()})


def register_events(event_bus: EventBus):
    """Subscribe to any events we need to support."""
    event_bus.subscribe(
        re.compile(f"^{REVOCATION_EVENT_PREFIX}{REVOCATION_REG_INIT_EVENT}.*"),
        on_revocation_registry_init_event,
    )
    event_bus.subscribe(
        re.compile(f"^{REVOCATION_EVENT_PREFIX}{REVOCATION_REG_ENDORSED_EVENT}.*"),
        on_revocation_registry_endorsed_event,
    )
    event_bus.subscribe(
        re.compile(f"^{REVOCATION_EVENT_PREFIX}{REVOCATION_ENTRY_EVENT}.*"),
        on_revocation_entry_event,
    )
    event_bus.subscribe(
        re.compile(f"^{REVOCATION_EVENT_PREFIX}REV_REG_ENTRY_TXN_FAILED.*"),
        on_rev_reg_entry_txn_failed,
    )


async def on_revocation_registry_init_event(profile: Profile, event: Event):
    """Handle revocation registry initiation event."""
    meta_data = event.payload
    if "endorser" in meta_data:
        # TODO error handling - for now just let exceptions get raised
        endorser_connection_id = meta_data["endorser"]["connection_id"]
        async with profile.session() as session:
            connection = await ConnRecord.retrieve_by_id(session, endorser_connection_id)
            endorser_info = await connection.metadata_get(session, "endorser_info")
        endorser_did = endorser_info["endorser_did"]
        write_ledger = False
    else:
        endorser_connection_id = None
        endorser_did = None
        write_ledger = True

    tails_base_url = profile.settings.get("tails_server_base_url")
    if not tails_base_url:
        raise RevocationError("tails_server_base_url not configured")

    # Generate the registry and upload the tails file
    async def generate(rr_record: IssuerRevRegRecord) -> dict:
        await rr_record.generate_registry(profile)
        public_uri = tails_base_url.rstrip("/") + f"/{registry_record.revoc_reg_id}"
        await rr_record.set_tails_file_public_uri(profile, public_uri)
        rev_reg_resp = await rr_record.send_def(
            profile,
            write_ledger=write_ledger,
            endorser_did=endorser_did,
        )
        if write_ledger:
            # Upload the tails file
            await rr_record.upload_tails_file(profile)

            # Post the initial revocation entry
            await notify_revocation_entry_event(profile, record_id, meta_data)
        else:
            transaction_manager = TransactionManager(profile)
            try:
                revo_transaction = await transaction_manager.create_record(
                    messages_attach=rev_reg_resp["result"],
                    connection_id=connection.connection_id,
                    meta_data=event.payload,
                )
            except StorageError as err:
                raise TransactionManagerError(reason=err.roll_up) from err

            # if auto-request, send the request to the endorser
            if profile.settings.get_value("endorser.auto_request"):
                try:
                    (
                        revo_transaction,
                        revo_transaction_request,
                    ) = await transaction_manager.create_request(
                        transaction=revo_transaction,
                        # TODO see if we need to parametrize these params
                        # expires_time=expires_time,
                    )
                except (StorageError, TransactionManagerError) as err:
                    raise TransactionManagerError(reason=err.roll_up) from err

                responder = profile.inject_or(BaseResponder)
                if responder:
                    await responder.send(
                        revo_transaction_request,
                        connection_id=connection.connection_id,
                    )
                else:
                    LOGGER.warning(
                        "Configuration has no BaseResponder: cannot update "
                        "revocation on registry ID: %s",
                        record_id,
                    )

    record_id = meta_data["context"]["issuer_rev_id"]
    async with profile.session() as session:
        registry_record = await IssuerRevRegRecord.retrieve_by_id(session, record_id)
    await shield(generate(registry_record))

    create_pending_rev_reg = meta_data["processing"].get("create_pending_rev_reg", False)
    if write_ledger and create_pending_rev_reg:
        revoc = IndyRevocation(profile)
        await revoc.init_issuer_registry(
            registry_record.cred_def_id,
            registry_record.max_cred_num,
            registry_record.revoc_def_type,
            endorser_connection_id=endorser_connection_id,
        )


async def on_revocation_entry_event(profile: Profile, event: Event):
    """Handle revocation entry event."""
    meta_data = event.payload
    if "endorser" in meta_data:
        # TODO error handling - for now just let exceptions get raised
        async with profile.session() as session:
            connection = await ConnRecord.retrieve_by_id(
                session, meta_data["endorser"]["connection_id"]
            )
            endorser_info = await connection.metadata_get(session, "endorser_info")
        endorser_did = endorser_info["endorser_did"]
        write_ledger = False
    else:
        endorser_did = None
        write_ledger = True

    record_id = meta_data["context"]["issuer_rev_id"]
    async with profile.session() as session:
        registry_record = await IssuerRevRegRecord.retrieve_by_id(session, record_id)
    rev_entry_resp = await registry_record.send_entry(
        profile,
        write_ledger=write_ledger,
        endorser_did=endorser_did,
    )

    if not write_ledger:
        transaction_manager = TransactionManager(profile)
        try:
            revo_transaction = await transaction_manager.create_record(
                messages_attach=rev_entry_resp["result"],
                connection_id=connection.connection_id,
                meta_data=meta_data,
            )
        except StorageError as err:
            raise RevocationError(err.roll_up) from err

        # if auto-request, send the request to the endorser
        if profile.settings.get_value("endorser.auto_request"):
            try:
                (
                    revo_transaction,
                    revo_transaction_request,
                ) = await transaction_manager.create_request(
                    transaction=revo_transaction,
                    # TODO see if we need to parametrize these params
                    # expires_time=expires_time,
                )
            except (StorageError, TransactionManagerError) as err:
                raise RevocationError(err.roll_up) from err

            responder = profile.inject_or(BaseResponder)
            if responder:
                await responder.send(
                    revo_transaction_request,
                    connection_id=connection.connection_id,
                )
            else:
                LOGGER.warning(
                    "Configuration has no BaseResponder: cannot update "
                    "revocation on cred def %s",
                    meta_data["endorser"]["cred_def_id"],
                )


async def on_revocation_registry_endorsed_event(profile: Profile, event: Event):
    """Handle revocation registry endorsement event."""
    meta_data = event.payload
    rev_reg_id = meta_data["context"]["rev_reg_id"]
    revoc = IndyRevocation(profile)
    registry_record = await revoc.get_issuer_rev_reg_record(rev_reg_id)

    if profile.settings.get_value("endorser.auto_request"):
        # NOTE: if there are multiple pods, then the one processing this
        # event may not be the one that generated the tails file.
        await registry_record.upload_tails_file(profile)

        # Post the initial revocation entry
        await notify_revocation_entry_event(profile, registry_record.record_id, meta_data)

    # create a "pending" registry if one is requested
    # (this is done automatically when creating a credential definition, so that when a
    #   revocation registry fills up, we can continue to issue credentials without a
    #   delay)
    create_pending_rev_reg = meta_data["processing"].get("create_pending_rev_reg", False)
    if create_pending_rev_reg:
        endorser_connection_id = (
            meta_data["endorser"].get("connection_id", None)
            if "endorser" in meta_data
            else None
        )
        await revoc.init_issuer_registry(
            registry_record.cred_def_id,
            registry_record.max_cred_num,
            registry_record.revoc_def_type,
            endorser_connection_id=endorser_connection_id,
        )


async def on_rev_reg_entry_txn_failed(profile: Profile, event: Event):
    """Handle revocation registry entry transaction failed event."""
    manager = RevocationManager(profile)
    await manager.fix_and_publish_from_invalid_accum_err(event.payload.get("msg"))


class TailsDeleteResponseSchema(OpenAPISchema):
    """Return schema for tails deletion."""

    message = fields.Str()


@querystring_schema(RevRegId())
@response_schema(TailsDeleteResponseSchema())
@docs(tags=["revocation"], summary="Delete the tail files")
@tenant_authentication
async def delete_tails(request: web.BaseRequest) -> json:
    """Delete Tails Files."""
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.query.get("rev_reg_id")
    cred_def_id = request.query.get("cred_def_id")
    revoc = IndyRevocation(profile)
    session = revoc._profile.session()
    if rev_reg_id:
        rev_reg = await revoc.get_issuer_rev_reg_record(rev_reg_id)
        tails_path = rev_reg.tails_local_path
        main_dir_rev = os.path.dirname(tails_path)
        try:
            shutil.rmtree(main_dir_rev)
            return web.json_response({"message": "All files deleted successfully"})
        except Exception as e:
            return web.json_response({"message": str(e)})
    elif cred_def_id:
        async with session:
            records = sorted(
                await IssuerRevRegRecord.query_by_cred_def_id(
                    session, cred_def_id, IssuerRevRegRecord.STATE_GENERATED
                )
            )
            if not records:
                return web.json_response(
                    {"message": "No tail files found for deletion"}, status=404
                )
            cred_reg = records[0]
        tails_path = cred_reg.tails_local_path
        main_dir_rev = os.path.dirname(tails_path)
        main_dir_cred = os.path.dirname(main_dir_rev)
        filenames = os.listdir(main_dir_cred)
        try:
            flag = 0
            for i in filenames:
                safe_cred_def_id = re.escape(cred_def_id)
                if re.search(safe_cred_def_id, i):
                    shutil.rmtree(main_dir_cred + "/" + i)
                    flag = 1
            if flag:
                return web.json_response({"message": "All files deleted successfully"})
            else:
                return web.json_response({"message": "No such file or directory"})

        except Exception as e:
            return web.json_response({"message": str(e)})


async def register(app: web.Application):
    """Register routes."""
    app.add_routes(
        [
            web.post("/revocation/revoke", revoke),
            web.post("/revocation/publish-revocations", publish_revocations),
            web.post(
                "/revocation/clear-pending-revocations",
                clear_pending_revocations,
            ),
            web.get(
                "/revocation/credential-record", get_cred_rev_record, allow_head=False
            ),
            web.get(
                "/revocation/registries/created",
                rev_regs_created,
                allow_head=False,
            ),
            web.get("/revocation/registry/{rev_reg_id}", get_rev_reg, allow_head=False),
            web.get(
                "/revocation/active-registry/{cred_def_id}",
                get_active_rev_reg,
                allow_head=False,
            ),
            web.post(
                "/revocation/active-registry/{cred_def_id}/rotate",
                rotate_rev_reg,
            ),
            web.get(
                "/revocation/registry/{rev_reg_id}/issued",
                get_rev_reg_issued_count,
                allow_head=False,
            ),
            web.get(
                "/revocation/registry/{rev_reg_id}/issued/details",
                get_rev_reg_issued,
                allow_head=False,
            ),
            web.get(
                "/revocation/registry/{rev_reg_id}/issued/indy_recs",
                get_rev_reg_indy_recs,
                allow_head=False,
            ),
            web.post("/revocation/create-registry", create_rev_reg),
            web.post("/revocation/registry/{rev_reg_id}/definition", send_rev_reg_def),
            web.post("/revocation/registry/{rev_reg_id}/entry", send_rev_reg_entry),
            web.patch("/revocation/registry/{rev_reg_id}", update_rev_reg),
            web.put("/revocation/registry/{rev_reg_id}/tails-file", upload_tails_file),
            web.get(
                "/revocation/registry/{rev_reg_id}/tails-file",
                get_tails_file,
                allow_head=False,
            ),
            web.patch(
                "/revocation/registry/{rev_reg_id}/set-state",
                set_rev_reg_state,
            ),
            web.put(
                "/revocation/registry/{rev_reg_id}/fix-revocation-entry-state",
                update_rev_reg_revoked_state,
            ),
            web.delete("/revocation/registry/delete-tails-file", delete_tails),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "revocation",
            "description": "Revocation registry management",
            "externalDocs": {
                "description": "Overview",
                "url": (
                    "https://github.com/hyperledger/indy-hipe/tree/"
                    "master/text/0011-cred-revocation"
                ),
            },
        }
    )

    # aio_http-apispec polite API only works on schema for JSON objects, not files yet
    methods = app._state["swagger_dict"]["paths"].get(
        "/revocation/registry/{rev_reg_id}/tails-file"
    )
    if methods:
        methods["get"]["responses"]["200"]["schema"] = {
            "type": "string",
            "format": "binary",
        }



import json
from base64 import urlsafe_b64decode
from unittest import IsolatedAsyncioTestCase

from ...resolver.did_resolver import DIDResolver
from ...resolver.tests.test_did_resolver import MockResolver
from ...utils.testing import create_test_profile
from ...wallet.did_method import KEY, DIDMethods
from ...wallet.key_type import ED25519, KeyTypes
from ..base import BaseWallet
from ..default_verification_key_strategy import (
    BaseVerificationKeyStrategy,
    DefaultVerificationKeyStrategy,
)
from ..jwt import jwt_sign
from ..sd_jwt import SDJWTVerifyResult, sd_jwt_sign, sd_jwt_verify


class TestSDJWT(IsolatedAsyncioTestCase):
    """Tests for JWT sign and verify using dids."""

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.resolver = DIDResolver()
        self.resolver.register_resolver(
            MockResolver(
                ["key"],
                resolved={
                    "@context": "https://www.w3.org/ns/did/v1",
                    "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                    "verificationMethod": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
                        }
                    ],
                    "authentication": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "assertionMethod": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "capabilityDelegation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "capabilityInvocation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "keyAgreement": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6LSbkodSr6SU2trs8VUgnrnWtSm7BAPG245ggrBmSrxbv1R",
                            "type": "X25519KeyAgreementKey2019",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "5dTvYHaNaB7mk7iA9LqCJEHG2dGZQsvoi8WGzDRtYEf",
                        }
                    ],
                },
                native=True,
            )
        )
        self.profile.context.injector.bind_instance(DIDMethods, DIDMethods())
        self.profile.context.injector.bind_instance(
            BaseVerificationKeyStrategy, DefaultVerificationKeyStrategy()
        )
        self.profile.context.injector.bind_instance(DIDResolver, self.resolver)
        self.profile.context.injector.bind_instance(KeyTypes, KeyTypes())

    seed = "testseed000000000000000000000001"
    headers = {}

    async def test_sign_with_did_key_and_verify(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            did_info = await wallet.create_local_did(KEY, ED25519, self.seed)

        verification_method = None
        payload = {
            "sub": "user_42",
            "given_name": "John",
            "family_name": "Doe",
            "email": "johndoe@example.com",
            "phone_number": "+1-202-555-0101",
            "phone_number_verified": True,
            "address": {
                "street_address": "123 Main St",
                "locality": "Anytown",
                "region": "Anystate",
                "country": "US",
            },
            "birthdate": "1940-01-01",
            "updated_at": 1570000000,
            "nationalities": ["US", "DE", "SA"],
            "iss": "https://example.com/issuer",
            "iat": 1683000000,
            "exp": 1883000000,
        }
        non_sd_list = [
            "given_name",
            "family_name",
            "birthdate",
        ]
        signed = await sd_jwt_sign(
            self.profile,
            self.headers,
            payload,
            non_sd_list,
            did_info.did,
            verification_method,
        )
        assert signed

        # Separate the jwt from the disclosures
        signed_sd_jwt = signed.split("~")[0]

        # Determine which selectively disclosable attributes to reveal
        revealed = ["sub", "phone_number", "phone_number_verified"]

        for disclosure in signed.split("~")[1:-1]:
            # Decode the disclosures
            padded = f"{disclosure}{'=' * divmod(len(disclosure), 4)[1]}"
            decoded = json.loads(urlsafe_b64decode(padded).decode("utf-8"))
            # Add the disclosures associated with the claims to be revealed
            if decoded[1] in revealed:
                signed_sd_jwt = signed_sd_jwt + "~" + disclosure

        verified = await sd_jwt_verify(self.profile, f"{signed_sd_jwt}~")
        assert verified.valid
        # Validate that the non-selectively disclosable claims are visible in the payload
        assert verified.payload["given_name"] == payload["given_name"]
        assert verified.payload["family_name"] == payload["family_name"]
        assert verified.payload["birthdate"] == payload["birthdate"]
        # Validate that the revealed claims are in the disclosures
        assert sorted(revealed) == sorted(
            [disclosure[1] for disclosure in verified.disclosures]
        )
        assert verified.payload["iss"] == payload["iss"]
        assert verified.payload["iat"] == payload["iat"]
        assert verified.payload["exp"] == payload["exp"]

    async def test_flat_structure(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            did_info = await wallet.create_local_did(KEY, ED25519, self.seed)

        verification_method = None
        non_sd_list = [
            "address.street_address",
            "address.locality",
            "address.region",
            "address.country",
        ]
        signed = await sd_jwt_sign(
            self.profile,
            self.headers,
            {
                "address": {
                    "street_address": "123 Main St",
                    "locality": "Anytown",
                    "region": "Anystate",
                    "country": "US",
                },
                "iss": "https://example.com/issuer",
                "iat": 1683000000,
                "exp": 1883000000,
            },
            non_sd_list,
            did_info.did,
            verification_method,
        )
        assert signed

        verified = await sd_jwt_verify(self.profile, signed)
        assert isinstance(verified, SDJWTVerifyResult)
        assert verified.valid
        assert verified.payload["_sd"]
        assert verified.payload["_sd_alg"]
        assert verified.disclosures[0][1] == "address"
        assert verified.disclosures[0][2] == {
            "street_address": "123 Main St",
            "locality": "Anytown",
            "region": "Anystate",
            "country": "US",
        }

    async def test_nested_structure(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            did_info = await wallet.create_local_did(KEY, ED25519, self.seed)
        verification_method = None
        non_sd_list = ["address"]

        signed = await sd_jwt_sign(
            self.profile,
            self.headers,
            {
                "address": {
                    "street_address": "123 Main St",
                    "locality": "Anytown",
                    "region": "Anystate",
                    "country": "US",
                },
                "iss": "https://example.com/issuer",
                "iat": 1683000000,
                "exp": 1883000000,
            },
            non_sd_list,
            did_info.did,
            verification_method,
        )
        assert signed

        verified = await sd_jwt_verify(self.profile, signed)
        assert isinstance(verified, SDJWTVerifyResult)
        assert verified.valid
        assert len(verified.payload["address"]["_sd"]) >= 4
        assert verified.payload["_sd_alg"]
        sd_claims = ["street_address", "region", "locality", "country"]
        assert sorted(sd_claims) == sorted([claim[1] for claim in verified.disclosures])

    async def test_recursive_nested_structure(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            did_info = await wallet.create_local_did(KEY, ED25519, self.seed)
        verification_method = None
        non_sd_list = []

        signed = await sd_jwt_sign(
            self.profile,
            self.headers,
            {
                "address": {
                    "street_address": "123 Main St",
                    "locality": "Anytown",
                    "region": "Anystate",
                    "country": "US",
                },
                "iss": "https://example.com/issuer",
                "iat": 1683000000,
                "exp": 1883000000,
            },
            non_sd_list,
            did_info.did,
            verification_method,
        )
        assert signed

        verified = await sd_jwt_verify(self.profile, signed)
        assert isinstance(verified, SDJWTVerifyResult)
        assert verified.valid
        assert "address" not in verified.payload
        assert verified.payload["_sd"]
        assert verified.payload["_sd_alg"]
        sd_claims = ["street_address", "region", "locality", "country"]
        for disclosure in verified.disclosures:
            if disclosure[1] == "address":
                assert isinstance(disclosure[2], dict)
                assert len(disclosure[2]["_sd"]) >= 4
            else:
                assert disclosure[1] in sd_claims

    async def test_list_splice(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            did_info = await wallet.create_local_did(KEY, ED25519, self.seed)
        payload = {"nationalities": ["US", "DE", "SA"]}
        verification_method = None
        non_sd_list = ["nationalities", "nationalities[1:3]"]

        signed = await sd_jwt_sign(
            self.profile,
            self.headers,
            payload,
            non_sd_list,
            did_info.did,
            verification_method,
        )
        assert signed

        verified = await sd_jwt_verify(self.profile, signed)
        assert isinstance(verified, SDJWTVerifyResult)
        assert verified.valid
        for nationality in verified.payload["nationalities"]:
            if isinstance(nationality, dict):
                assert nationality["..."]
                assert len(nationality) == 1
            else:
                assert nationality in payload["nationalities"]
        assert verified.payload["_sd_alg"]
        assert verified.disclosures[0][1] == "US"

    async def test_sd_jwt_key_binding(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            did_info = await wallet.create_local_did(KEY, ED25519, self.seed)
        verification_method = None

        payload = {
            "given_name": "John",
            "family_name": "Doe",
            "iss": "https://example.com/issuer",
            "iat": 1683000000,
            "exp": 1883000000,
            "cnf": {
                "jwk": {
                    "kty": "EC",
                    "crv": "P-256",
                    "x": "TCAER19Zvu3OHF4j4W4vfSVoHIP1ILilDls7vCeGemc",
                    "y": "ZxjiWWbZMQGHVWKVQ4hbSIirsVfuecCE6t4jT9F2HZQ",
                }
            },
        }
        signed = await sd_jwt_sign(
            self.profile,
            self.headers,
            payload,
            did=did_info.did,
            verification_method=verification_method,
        )
        assert signed

        # Key binding
        headers_kb = {"alg": "ES256", "typ": "kb+jwt"}
        payload_kb = {
            "nonce": "1234567890",
            "aud": "https://example.com/verifier",
            "iat": 1688160483,
        }
        signed_kb = await jwt_sign(
            self.profile,
            headers_kb,
            payload_kb,
            did_info.did,
            verification_method,
        )
        assert signed_kb

        assert await sd_jwt_verify(
            self.profile,
            f"{signed}{signed_kb}",
            expected_aud=payload_kb["aud"],
            expected_nonce=payload_kb["nonce"],
        )

    test_input = [
        (
            "Either both expected_aud and expected_nonce must be provided or both must be None",
            {
                "given_name": "John",
                "family_name": "Doe",
                "iss": "https://example.com/issuer",
                "iat": 1683000000,
                "exp": 1883000000,
                "cnf": {
                    "jwk": {
                        "kty": "EC",
                        "crv": "P-256",
                        "x": "TCAER19Zvu3OHF4j4W4vfSVoHIP1ILilDls7vCeGemc",
                        "y": "ZxjiWWbZMQGHVWKVQ4hbSIirsVfuecCE6t4jT9F2HZQ",
                    }
                },
            },
            {"alg": "ES256", "typ": "kb+jwt"},
            "https://example.com/verifier",
            None,
        ),
        (
            "No holder public key in SD-JWT",
            {
                "given_name": "John",
                "family_name": "Doe",
                "iss": "https://example.com/issuer",
                "iat": 1683000000,
                "exp": 1883000000,
            },
            {"alg": "ES256", "typ": "kb+jwt"},
            "https://example.com/verifier",
            "1234567890",
        ),
        (
            "The holder_public_key_payload is malformed. It doesn't contain the claim jwk: ",
            {
                "given_name": "John",
                "family_name": "Doe",
                "iss": "https://example.com/issuer",
                "iat": 1683000000,
                "exp": 1883000000,
                "cnf": {"y": "ZxjiWWbZMQGHVWKVQ4hbSIirsVfuecCE6t4jT9F2HZQ"},
            },
            {"alg": "ES256", "typ": "kb+jwt"},
            "https://example.com/verifier",
            "1234567890",
        ),
        (
            "Invalid header typ",
            {
                "given_name": "John",
                "family_name": "Doe",
                "iss": "https://example.com/issuer",
                "iat": 1683000000,
                "exp": 1883000000,
                "cnf": {
                    "jwk": {
                        "kty": "EC",
                        "crv": "P-256",
                        "x": "TCAER19Zvu3OHF4j4W4vfSVoHIP1ILilDls7vCeGemc",
                        "y": "ZxjiWWbZMQGHVWKVQ4hbSIirsVfuecCE6t4jT9F2HZQ",
                    }
                },
            },
            {"alg": "ES256", "typ": "JWT"},
            "https://example.com/verifier",
            "1234567890",
        ),
        (
            "Invalid audience",
            {
                "given_name": "John",
                "family_name": "Doe",
                "iss": "https://example.com/issuer",
                "iat": 1683000000,
                "exp": 1883000000,
                "cnf": {
                    "jwk": {
                        "kty": "EC",
                        "crv": "P-256",
                        "x": "TCAER19Zvu3OHF4j4W4vfSVoHIP1ILilDls7vCeGemc",
                        "y": "ZxjiWWbZMQGHVWKVQ4hbSIirsVfuecCE6t4jT9F2HZQ",
                    }
                },
            },
            {"alg": "ES256", "typ": "kb+jwt"},
            "invalid_aud",
            "1234567890",
        ),
        (
            "Invalid nonce",
            {
                "given_name": "John",
                "family_name": "Doe",
                "iss": "https://example.com/issuer",
                "iat": 1683000000,
                "exp": 1883000000,
                "cnf": {
                    "jwk": {
                        "kty": "EC",
                        "crv": "P-256",
                        "x": "TCAER19Zvu3OHF4j4W4vfSVoHIP1ILilDls7vCeGemc",
                        "y": "ZxjiWWbZMQGHVWKVQ4hbSIirsVfuecCE6t4jT9F2HZQ",
                    }
                },
            },
            {"alg": "ES256", "typ": "kb+jwt"},
            "https://example.com/verifier",
            "invalid_nonce",
        ),
    ]

    async def test_sd_jwt_key_binding_errors(self):
        for error, payload, headers_kb, expected_aud, expected_nonce in self.test_input:
            async with self.profile.session() as session:
                wallet = session.inject(BaseWallet)
                did_info = await wallet.create_local_did(KEY, ED25519, self.seed)
            verification_method = None

            signed = await sd_jwt_sign(
                self.profile,
                self.headers,
                payload,
                did=did_info.did,
                verification_method=verification_method,
            )
            assert signed

            # Key binding
            payload_kb = {
                "nonce": "1234567890",
                "aud": "https://example.com/verifier",
                "iat": 1688160483,
            }
            signed_kb = await jwt_sign(
                self.profile,
                headers_kb,
                payload_kb,
                did_info.did,
                verification_method,
            )
            assert signed_kb

        with self.assertRaises(
            ValueError,
        ):
            await sd_jwt_verify(
                self.profile,
                f"{signed}{signed_kb}",
                expected_aud=expected_aud,
                expected_nonce=expected_nonce,
            )



"""V2.0 issue-credential protocol manager."""

import logging
from typing import Mapping, Optional, Tuple

from ....connections.models.conn_record import ConnRecord
from ....core.error import BaseError
from ....core.oob_processor import OobRecord
from ....core.profile import Profile
from ....messaging.responder import BaseResponder
from ....storage.error import StorageError, StorageNotFoundError
from .messages.cred_ack import V20CredAck
from .messages.cred_format import V20CredFormat
from .messages.cred_issue import V20CredIssue
from .messages.cred_offer import V20CredOffer
from .messages.cred_problem_report import ProblemReportReason, V20CredProblemReport
from .messages.cred_proposal import V20CredProposal
from .messages.cred_request import V20CredRequest
from .messages.inner.cred_preview import V20CredPreview
from .models.cred_ex_record import V20CredExRecord

LOGGER = logging.getLogger(__name__)


class V20CredManagerError(BaseError):
    """Credential manager error under issue-credential protocol v2.0."""


class V20CredManager:
    """Class for managing credentials."""

    def __init__(self, profile: Profile):
        """Initialize a V20CredManager.

        Args:
            profile: The profile instance for this credential manager
        """
        self._profile = profile

    @property
    def profile(self) -> Profile:
        """Accessor for the current profile instance.

        Returns:
            The profile instance for this credential manager

        """
        return self._profile

    async def prepare_send(
        self,
        connection_id: str,
        cred_proposal: V20CredProposal,
        verification_method: Optional[str] = None,
        auto_remove: Optional[bool] = None,
        replacement_id: Optional[str] = None,
    ) -> Tuple[V20CredExRecord, V20CredOffer]:
        """Set up a new credential exchange record for an automated send.

        Args:
            connection_id: connection for which to create offer
            cred_proposal: credential proposal with preview
            verification_method: an optional verification method to be used when issuing
            auto_remove: flag to remove the record automatically on completion
            replacement_id: identifier to help coordinate credential replacement

        Returns:
            A tuple of the new credential exchange record and credential offer message

        """
        if auto_remove is None:
            auto_remove = not self._profile.settings.get("preserve_exchange_records")
        cred_ex_record = V20CredExRecord(
            connection_id=connection_id,
            verification_method=verification_method,
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            cred_proposal=cred_proposal,
            auto_issue=True,
            auto_remove=auto_remove,
            trace=(cred_proposal._trace is not None),
        )
        return await self.create_offer(
            cred_ex_record=cred_ex_record,
            counter_proposal=None,
            comment="create automated v2.0 credential exchange record",
            replacement_id=replacement_id,
        )

    async def create_proposal(
        self,
        connection_id: str,
        *,
        auto_remove: Optional[bool] = None,
        comment: Optional[str] = None,
        cred_preview: V20CredPreview,
        fmt2filter: Mapping[V20CredFormat.Format, Mapping[str, str]],
        trace: bool = False,
    ) -> V20CredExRecord:
        """Create a credential proposal.

        Args:
            connection_id: connection for which to create proposal
            auto_remove: whether to remove record automatically on completion
            comment: optional human-readable comment to include in proposal
            cred_preview: credential preview to use to create credential proposal
            fmt2filter: mapping between format and filter
            trace: whether to trace the operation

        Returns:
            Resulting credential exchange record including credential proposal

        """

        if auto_remove is None:
            auto_remove = not self._profile.settings.get("preserve_exchange_records")
        cred_ex_record = V20CredExRecord(
            connection_id=connection_id,
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_HOLDER,
            state=V20CredExRecord.STATE_PROPOSAL_SENT,
            auto_remove=auto_remove,
            trace=trace,
        )

        # Format specific create_proposal handler
        formats = [
            await fmt.handler(self._profile).create_proposal(cred_ex_record, filter)
            for (fmt, filter) in fmt2filter.items()
        ]

        cred_proposal_message = V20CredProposal(
            comment=comment,
            credential_preview=cred_preview,
            formats=[format for (format, _) in formats],
            filters_attach=[attach for (_, attach) in formats],
        )

        cred_ex_record.thread_id = cred_proposal_message._thread_id
        cred_ex_record.cred_proposal = cred_proposal_message

        cred_proposal_message.assign_trace_decorator(self._profile.settings, trace)

        async with self._profile.session() as session:
            await cred_ex_record.save(
                session,
                reason="create v2.0 credential proposal",
            )
        return cred_ex_record

    async def receive_proposal(
        self,
        cred_proposal_message: V20CredProposal,
        connection_id: str,
    ) -> V20CredExRecord:
        """Receive a credential proposal.

        Returns:
            The resulting credential exchange record, created

        """
        # at this point, cred def and schema still open to potential negotiation
        cred_ex_record = V20CredExRecord(
            connection_id=connection_id,
            thread_id=cred_proposal_message._thread_id,
            initiator=V20CredExRecord.INITIATOR_EXTERNAL,
            role=V20CredExRecord.ROLE_ISSUER,
            auto_offer=self._profile.settings.get(
                "debug.auto_respond_credential_proposal"
            ),
            auto_issue=self._profile.settings.get(
                "debug.auto_respond_credential_request"
            ),
            auto_remove=not self._profile.settings.get("preserve_exchange_records"),
            trace=(cred_proposal_message._trace is not None),
        )

        # Format specific receive_proposal handlers
        for format in cred_proposal_message.formats:
            await (
                V20CredFormat.Format.get(format.format)
                .handler(self.profile)
                .receive_proposal(cred_ex_record, cred_proposal_message)
            )

        cred_ex_record.cred_proposal = cred_proposal_message
        cred_ex_record.state = V20CredExRecord.STATE_PROPOSAL_RECEIVED

        async with self._profile.session() as session:
            await cred_ex_record.save(
                session,
                reason="receive v2.0 credential proposal",
            )

        return cred_ex_record

    async def create_offer(
        self,
        cred_ex_record: V20CredExRecord,
        counter_proposal: Optional[V20CredProposal] = None,
        replacement_id: Optional[str] = None,
        comment: Optional[str] = None,
    ) -> Tuple[V20CredExRecord, V20CredOffer]:
        """Create credential offer, update credential exchange record.

        Args:
            cred_ex_record (V20CredExRecord): The credential exchange record for
                which to create the offer.
            counter_proposal (V20CredProposal, optional): The counter proposal for
                the credential offer. Defaults to None.
            replacement_id (str, optional): Identifier to help coordinate credential
                replacement. Defaults to None.
            comment (str, optional): Optional human-readable comment to set in the
                offer message. Defaults to None.

        Returns:
            Tuple[V20CredExRecord, V20CredOffer]: A tuple containing the updated
                credential exchange record and the credential offer message.

        Raises:
            V20CredManagerError: If unable to create the credential offer due to no
                supported formats.

        """

        cred_proposal_message = (
            counter_proposal if counter_proposal else cred_ex_record.cred_proposal
        )
        cred_proposal_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        formats = []
        # Format specific create_offer handler
        for format in cred_proposal_message.formats:
            cred_format = V20CredFormat.Format.get(format.format)

            if cred_format:
                formats.append(
                    await cred_format.handler(self.profile).create_offer(
                        cred_proposal_message
                    )
                )

        if len(formats) == 0:
            raise V20CredManagerError(
                "Unable to create credential offer. No supported formats"
            )

        cred_offer_message = V20CredOffer(
            replacement_id=replacement_id,
            comment=comment,
            credential_preview=cred_proposal_message.credential_preview,
            formats=[format for (format, _) in formats],
            offers_attach=[attach for (_, attach) in formats],
        )

        cred_offer_message._thread = {"thid": cred_ex_record.thread_id}
        cred_offer_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        cred_ex_record.thread_id = cred_offer_message._thread_id
        cred_ex_record.state = V20CredExRecord.STATE_OFFER_SENT
        cred_ex_record.cred_proposal = (  # any counter replaces original
            cred_proposal_message
        )
        cred_ex_record.cred_offer = cred_offer_message

        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="create v2.0 credential offer")

        return (cred_ex_record, cred_offer_message)

    async def receive_offer(
        self,
        cred_offer_message: V20CredOffer,
        connection_id: Optional[str],
    ) -> V20CredExRecord:
        """Receive a credential offer.

        Args:
            cred_offer_message: credential offer message
            connection_id: connection identifier

        Returns:
            The credential exchange record, updated

        """

        # Get credential exchange record (holder sent proposal first)
        # or create it (issuer sent offer first)
        try:
            async with self._profile.session() as session:
                cred_ex_record = await V20CredExRecord.retrieve_by_conn_and_thread(
                    session,
                    connection_id,
                    cred_offer_message._thread_id,
                    role=V20CredExRecord.ROLE_HOLDER,
                )
        except StorageNotFoundError:  # issuer sent this offer free of any proposal
            cred_ex_record = V20CredExRecord(
                connection_id=connection_id,
                thread_id=cred_offer_message._thread_id,
                initiator=V20CredExRecord.INITIATOR_EXTERNAL,
                role=V20CredExRecord.ROLE_HOLDER,
                auto_remove=not self._profile.settings.get("preserve_exchange_records"),
                trace=(cred_offer_message._trace is not None),
            )

        # Format specific receive_offer handler
        for format in cred_offer_message.formats:
            cred_format = V20CredFormat.Format.get(format.format)

            if cred_format:
                await cred_format.handler(self.profile).receive_offer(
                    cred_ex_record, cred_offer_message
                )

        cred_ex_record.cred_offer = cred_offer_message
        cred_ex_record.state = V20CredExRecord.STATE_OFFER_RECEIVED

        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="receive v2.0 credential offer")

        return cred_ex_record

    async def create_request(
        self,
        cred_ex_record: V20CredExRecord,
        holder_did: str,
        comment: Optional[str] = None,
    ) -> Tuple[V20CredExRecord, V20CredRequest]:
        """Create a credential request.

        Args:
            cred_ex_record: credential exchange record for which to create request
            holder_did: holder DID
            comment: optional human-readable comment to set in request message

        Returns:
            A tuple (credential exchange record, credential request message)

        """
        if cred_ex_record.cred_request:
            raise V20CredManagerError(
                "create_request() called multiple times for "
                f"v2.0 credential exchange {cred_ex_record.cred_ex_id}"
            )

        # react to credential offer, use offer formats
        if cred_ex_record.state:
            if cred_ex_record.state != V20CredExRecord.STATE_OFFER_RECEIVED:
                raise V20CredManagerError(
                    f"Credential exchange {cred_ex_record.cred_ex_id} "
                    f"in {cred_ex_record.state} state "
                    f"(must be {V20CredExRecord.STATE_OFFER_RECEIVED})"
                )

            cred_offer = cred_ex_record.cred_offer

            input_formats = cred_offer.formats
        # start with request (not allowed for indy -> checked in indy format handler)
        # use proposal formats
        else:
            cred_proposal = cred_ex_record.cred_proposal
            input_formats = cred_proposal.formats

        request_formats = []
        # Format specific create_request handler
        for format in input_formats:
            cred_format = V20CredFormat.Format.get(format.format)

            if cred_format:
                request_formats.append(
                    await cred_format.handler(self.profile).create_request(
                        cred_ex_record, {"holder_did": holder_did}
                    )
                )

        if len(request_formats) == 0:
            raise V20CredManagerError(
                "Unable to create credential request. No supported formats"
            )

        cred_request_message = V20CredRequest(
            comment=comment,
            formats=[format for (format, _) in request_formats],
            requests_attach=[attach for (_, attach) in request_formats],
        )

        # Assign thid (and optionally pthid) to message
        cred_request_message.assign_thread_from(cred_ex_record.cred_offer)
        cred_request_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        cred_ex_record.thread_id = cred_request_message._thread_id
        cred_ex_record.state = V20CredExRecord.STATE_REQUEST_SENT
        cred_ex_record.cred_request = cred_request_message

        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="create v2.0 credential request")

        return (cred_ex_record, cred_request_message)

    async def receive_request(
        self,
        cred_request_message: V20CredRequest,
        connection_record: Optional[ConnRecord],
        oob_record: Optional[OobRecord],
    ) -> V20CredExRecord:
        """Receive a credential request.

        This method is called when a credential request is received from a connection.
        It processes the request, updates the credential exchange record, and returns
        the updated record.

        Args:
            cred_request_message: The credential request to receive.
            connection_record: The connection record associated with the request.
            oob_record: The out-of-band record associated with the request.

        Returns:
            The updated credential exchange record.

        Raises:
            StorageNotFoundError: If the credential exchange record is not found.

        """
        # connection_id is None in the record if this is in response to
        # an request~attach from an OOB message. If so, we do not want to filter
        # the record by connection_id.
        connection_id = None if oob_record else connection_record.connection_id

        handlers = [
            f.handler(self.profile)
            for format in cred_request_message.formats
            if (f := V20CredFormat.Format.get(format.format))
        ]
        handlers_without_offer = [
            handler for handler in handlers if handler.can_receive_request_without_offer()
        ]

        async with self._profile.session() as session:
            try:
                cred_ex_record = await V20CredExRecord.retrieve_by_conn_and_thread(
                    session,
                    connection_id,
                    cred_request_message._thread_id,
                    role=V20CredExRecord.ROLE_ISSUER,
                )
            except StorageNotFoundError as ex:
                # holder sent this request free of any offer
                if handlers_without_offer:
                    handlers = handlers_without_offer
                    cred_ex_record = V20CredExRecord(
                        connection_id=connection_id,
                        thread_id=cred_request_message._thread_id,
                        initiator=V20CredExRecord.INITIATOR_EXTERNAL,
                        role=V20CredExRecord.ROLE_ISSUER,
                        auto_remove=not self._profile.settings.get(
                            "preserve_exchange_records"
                        ),
                        trace=(cred_request_message._trace is not None),
                        auto_issue=self._profile.settings.get(
                            "debug.auto_respond_credential_request"
                        ),
                    )
                else:
                    raise ex

        if connection_record:
            cred_ex_record.connection_id = connection_record.connection_id

        for handler in handlers:
            await handler.receive_request(cred_ex_record, cred_request_message)

        cred_ex_record.cred_request = cred_request_message
        cred_ex_record.state = V20CredExRecord.STATE_REQUEST_RECEIVED

        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="receive v2.0 credential request")

        return cred_ex_record

    async def issue_credential(
        self,
        cred_ex_record: V20CredExRecord,
        *,
        comment: Optional[str] = None,
    ) -> Tuple[V20CredExRecord, V20CredIssue]:
        """Issue a credential.

        Args:
            cred_ex_record: credential exchange record for which to issue credential
            comment: optional human-readable comment pertaining to credential issue

        Returns:
            Tuple: (Updated credential exchange record, credential issue message)

        """

        if cred_ex_record.state != V20CredExRecord.STATE_REQUEST_RECEIVED:
            raise V20CredManagerError(
                f"Credential exchange {cred_ex_record.cred_ex_id} "
                f"in {cred_ex_record.state} state "
                f"(must be {V20CredExRecord.STATE_REQUEST_RECEIVED})"
            )

        if cred_ex_record.cred_issue:
            raise V20CredManagerError(
                "issue_credential() called multiple times for "
                f"cred ex record {cred_ex_record.cred_ex_id}"
            )

        replacement_id = None
        input_formats = cred_ex_record.cred_request.formats

        if cred_ex_record.cred_offer:
            cred_offer_message = cred_ex_record.cred_offer
            replacement_id = cred_offer_message.replacement_id

        # Format specific issue_credential handler
        issue_formats = []
        for format in input_formats:
            cred_format = V20CredFormat.Format.get(format.format)

            if cred_format:
                issue_formats.append(
                    await cred_format.handler(self.profile).issue_credential(
                        cred_ex_record
                    )
                )

        if len(issue_formats) == 0:
            raise V20CredManagerError("Unable to issue credential. No supported formats")

        cred_issue_message = V20CredIssue(
            replacement_id=replacement_id,
            comment=comment,
            formats=[format for (format, _) in issue_formats],
            credentials_attach=[attach for (_, attach) in issue_formats],
        )

        cred_ex_record.state = V20CredExRecord.STATE_ISSUED
        cred_ex_record.cred_issue = cred_issue_message
        async with self._profile.session() as session:
            # FIXME - re-fetch record to check state, apply transactional update
            await cred_ex_record.save(session, reason="v2.0 issue credential")

        cred_issue_message._thread = {"thid": cred_ex_record.thread_id}
        cred_issue_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        return (cred_ex_record, cred_issue_message)

    async def receive_credential(
        self, cred_issue_message: V20CredIssue, connection_id: Optional[str]
    ) -> V20CredExRecord:
        """Receive a credential issue message from an issuer.

        Hold cred in storage potentially to be processed by controller before storing.

        Returns:
            Credential exchange record, retrieved and updated

        """
        assert cred_issue_message.credentials_attach

        # FIXME use transaction, fetch for_update
        async with self._profile.session() as session:
            cred_ex_record = await V20CredExRecord.retrieve_by_conn_and_thread(
                session,
                connection_id,
                cred_issue_message._thread_id,
                role=V20CredExRecord.ROLE_HOLDER,
            )

        cred_request_message = cred_ex_record.cred_request
        req_formats = [
            V20CredFormat.Format.get(fmt.format)
            for fmt in cred_request_message.formats
            if V20CredFormat.Format.get(fmt.format)
        ]
        issue_formats = [
            V20CredFormat.Format.get(fmt.format)
            for fmt in cred_issue_message.formats
            if V20CredFormat.Format.get(fmt.format)
        ]
        handled_formats = []

        def _check_formats():
            """Allow indy issue fomat and anoncreds req format or matching formats."""
            return (
                issue_formats == [V20CredFormat.Format.INDY]
                and req_formats == [V20CredFormat.Format.ANONCREDS]
            ) or len(set(issue_formats) - set(req_formats)) == 0

        if not _check_formats():
            raise V20CredManagerError(
                "Received issue credential format(s) not present in credential "
                f"request: {set(issue_formats) - set(req_formats)}"
            )

        for issue_format in issue_formats:
            await issue_format.handler(self.profile).receive_credential(
                cred_ex_record, cred_issue_message
            )
            handled_formats.append(issue_format)

        if len(handled_formats) == 0:
            raise V20CredManagerError("No supported credential formats received.")

        cred_ex_record.cred_issue = cred_issue_message
        cred_ex_record.state = V20CredExRecord.STATE_CREDENTIAL_RECEIVED

        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="receive v2.0 credential issue")
        return cred_ex_record

    async def store_credential(
        self, cred_ex_record: V20CredExRecord, cred_id: Optional[str] = None
    ) -> Tuple[V20CredExRecord, V20CredAck]:
        """Store a credential in holder wallet; send ack to issuer.

        Args:
            cred_ex_record: credential exchange record with credential to store and ack
            cred_id: optional credential identifier to override default on storage

        Returns:
            Updated credential exchange record

        """
        if cred_ex_record.state != (V20CredExRecord.STATE_CREDENTIAL_RECEIVED):
            raise V20CredManagerError(
                f"Credential exchange {cred_ex_record.cred_ex_id} "
                f"in {cred_ex_record.state} state "
                f"(must be {V20CredExRecord.STATE_CREDENTIAL_RECEIVED})"
            )

        # Format specific store_credential handler
        for format in cred_ex_record.cred_issue.formats:
            cred_format = V20CredFormat.Format.get(format.format)

            if cred_format:
                await cred_format.handler(self.profile).store_credential(
                    cred_ex_record, cred_id
                )
                # TODO: if storing multiple credentials we can't reuse the same id
                cred_id = None

        return cred_ex_record

    async def send_cred_ack(
        self,
        cred_ex_record: V20CredExRecord,
    ):
        """Create, send, and return ack message for input cred ex record.

        Delete cred ex record if set to auto-remove.

        Returns:
            Tuple: cred ex record, cred ack message for tracing

        """
        cred_ack_message = V20CredAck()
        cred_ack_message.assign_thread_id(
            cred_ex_record.thread_id, cred_ex_record.parent_thread_id
        )
        cred_ack_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        cred_ex_record.state = V20CredExRecord.STATE_DONE
        try:
            async with self._profile.session() as session:
                # FIXME - re-fetch record to check state, apply transactional update
                await cred_ex_record.save(session, reason="store credential v2.0")

            if cred_ex_record.auto_remove:
                await self.delete_cred_ex_record(cred_ex_record.cred_ex_id)

        except StorageError:
            LOGGER.exception(
                "Error sending credential ack"
            )  # holder still owes an ack: carry on

        responder = self._profile.inject_or(BaseResponder)
        if responder:
            await responder.send_reply(
                cred_ack_message,
                connection_id=cred_ex_record.connection_id,
            )
        else:
            LOGGER.warning(
                "Configuration has no BaseResponder: cannot ack credential on %s",
                cred_ex_record.thread_id,
            )

        return cred_ex_record, cred_ack_message

    async def receive_credential_ack(
        self, cred_ack_message: V20CredAck, connection_id: Optional[str]
    ) -> V20CredExRecord:
        """Receive credential ack from holder.

        Args:
            cred_ack_message: credential ack message to receive
            connection_id: connection identifier

        Returns:
            credential exchange record, retrieved and updated

        """
        # FIXME use transaction, fetch for_update
        async with self._profile.session() as session:
            cred_ex_record = await V20CredExRecord.retrieve_by_conn_and_thread(
                session,
                connection_id,
                cred_ack_message._thread_id,
                role=V20CredExRecord.ROLE_ISSUER,
            )

            cred_ex_record.state = V20CredExRecord.STATE_DONE
            await cred_ex_record.save(session, reason="receive credential ack v2.0")

        if cred_ex_record.auto_remove:
            await self.delete_cred_ex_record(cred_ex_record.cred_ex_id)

        return cred_ex_record

    async def delete_cred_ex_record(self, cred_ex_id: str) -> None:
        """Delete credential exchange record and associated detail records."""

        async with self._profile.session() as session:
            for fmt in V20CredFormat.Format:  # details first: do not strand any orphans
                for record in await fmt.detail.query_by_cred_ex_id(
                    session,
                    cred_ex_id,
                ):
                    await record.delete_record(session)

            cred_ex_record = await V20CredExRecord.retrieve_by_id(session, cred_ex_id)
            await cred_ex_record.delete_record(session)

    async def receive_problem_report(
        self, message: V20CredProblemReport, connection_id: str
    ):
        """Receive problem report.

        Returns:
            credential exchange record, retrieved and updated

        """
        # FIXME use transaction, fetch for_update
        async with self._profile.session() as session:
            cred_ex_record = await V20CredExRecord.retrieve_by_conn_and_thread(
                session,
                connection_id,
                message._thread_id,
            )

            cred_ex_record.state = V20CredExRecord.STATE_ABANDONED
            code = message.description.get(
                "code",
                ProblemReportReason.ISSUANCE_ABANDONED.value,
            )
            cred_ex_record.error_msg = f"{code}: {message.description.get('en', code)}"
            await cred_ex_record.save(session, reason="received problem report")

        return cred_ex_record



"""Connection request handler under RFC 23 (DID exchange)."""

from .....connections.models.conn_record import ConnRecord
from .....messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from ....coordinate_mediation.v1_0.manager import MediationManager
from ..manager import DIDXManager, DIDXManagerError
from ..messages.request import DIDXRequest


class DIDXRequestHandler(BaseHandler):
    """Handler class for connection request message under RFC 23 (DID exchange)."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle connection request under RFC 23 (DID exchange).

        Args:
            context: Request context
            responder: Responder callback
        """

        self._logger.debug("DIDXRequestHandler called with context %s", context)
        assert isinstance(context.message, DIDXRequest)

        profile = context.profile
        mgr = DIDXManager(profile)

        mediation_id = None
        if context.connection_record:
            async with profile.session() as session:
                mediation_metadata = await context.connection_record.metadata_get(
                    session, MediationManager.METADATA_KEY, {}
                )
            mediation_id = mediation_metadata.get(MediationManager.METADATA_ID)

        try:
            conn_rec = await mgr.receive_request(
                request=context.message,
                recipient_did=context.message_receipt.recipient_did,
                recipient_verkey=(
                    None
                    if context.message_receipt.recipient_did_public
                    else context.message_receipt.recipient_verkey
                ),
            )
            # Auto respond
            if conn_rec.accept == ConnRecord.ACCEPT_AUTO:
                response = await mgr.create_response(
                    conn_rec,
                    mediation_id=mediation_id,
                )
                await responder.send_reply(response, connection_id=conn_rec.connection_id)
                conn_rec.state = ConnRecord.State.RESPONSE.rfc23
                async with context.session() as session:
                    await conn_rec.save(session, reason="Sent connection response")
            else:
                self._logger.debug("DID exchange request will await acceptance")

        except DIDXManagerError as e:
            report, targets = await mgr.manager_error_to_problem_report(
                e, context.message, context.message_receipt
            )
            if report and targets:
                await responder.send_reply(
                    message=report,
                    target_list=targets,
                )



"""JSON-LD document loader methods."""

import asyncio
import concurrent.futures
from typing import Callable

import nest_asyncio
from pydid.did_url import DIDUrl
from pyld.documentloader import requests

from ...cache.base import BaseCache
from ...core.profile import Profile
from ...resolver.did_resolver import DIDResolver
from .document_downloader import StaticCacheJsonLdDownloader
from .error import LinkedDataProofException

nest_asyncio.apply()


class DocumentLoader:
    """JSON-LD document loader."""

    def __init__(self, profile: Profile, cache_ttl: int = 300) -> None:
        """Initialize new DocumentLoader instance.

        Args:
            profile (Profile): The profile
            cache_ttl (int, optional): TTL for cached documents. Defaults to 300.

        """
        self.profile = profile
        self.resolver = profile.inject(DIDResolver)
        self.cache = profile.inject_or(BaseCache)
        self.online_request_loader = requests.requests_document_loader()
        self.requests_loader = StaticCacheJsonLdDownloader().load
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self.cache_ttl = cache_ttl
        self._event_loop = asyncio.get_event_loop()

    async def _load_did_document(self, did: str, options: dict):
        # Resolver expects plain did without path, query, etc...
        # DIDUrl throws error if it contains no path, query etc...
        # This makes sure we get a plain did
        did = DIDUrl.parse(did).did if DIDUrl.is_valid(did) else did

        did_document = await self.resolver.resolve(self.profile, did)

        document = {
            "contentType": "application/ld+json",
            "contextUrl": None,
            "documentUrl": did,
            "document": did_document,
        }

        return document

    def _load_http_document(self, url: str, options: dict):
        document = self.requests_loader(url, options)

        return document

    # Async document loader can use await for cache and did resolver
    async def _load_async(self, url: str, options: dict):
        """Retrieve http(s) or did document."""

        # Resolve DIDs using did resolver
        if url.startswith("did:"):
            document = await self._load_did_document(url, options)
        elif url.startswith("http://") or url.startswith("https://"):
            document = self._load_http_document(url, options)
        else:
            raise LinkedDataProofException(
                "Unrecognized url format. Must start with 'did:', 'http://' or 'https://'"
            )

        return document

    async def load_document(self, url: str, options: dict):
        """Load JSON-LD document.

        Method signature conforms to PyLD document loader interface

        Document loading is processed in separate thread to deal with
        async to sync transformation.
        """
        cache_key = f"json_ld_document_resolver::{url}"

        # Try to get from cache
        if self.cache:
            document = await self.cache.get(cache_key)
            if document:
                return document

        document = await self._load_async(url, options)

        # Cache document, if cache is available
        if self.cache:
            await self.cache.set(cache_key, document, self.cache_ttl)

        return document

    def __call__(self, url: str, options: dict):
        """Load JSON-LD Document."""

        loop = self._event_loop
        coroutine = self.load_document(url, options)
        document = loop.run_until_complete(coroutine)

        return document


DocumentLoaderMethod = Callable[[str, dict], dict]

__all__ = ["DocumentLoaderMethod", "DocumentLoader"]



from unittest import IsolatedAsyncioTestCase
from unittest.mock import ANY

from ...admin.request_context import AdminRequestContext
from ...cache.base import BaseCache
from ...cache.in_memory import InMemoryCache
from ...connections.models.conn_record import ConnRecord
from ...storage.error import StorageNotFoundError
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import routes as test_module


class TestConnectionRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    async def test_connections_list(self):
        self.request.query = {
            "invitation_id": "dummy",  # exercise tag filter assignment
            "their_role": ConnRecord.Role.REQUESTER.rfc160,
            "connection_protocol": "connections/1.0",
            "invitation_key": "some-invitation-key",
            "their_public_did": "a_public_did",
            "invitation_msg_id": "dummy_msg",
        }

        STATE_COMPLETED = ConnRecord.State.COMPLETED
        STATE_INVITATION = ConnRecord.State.INVITATION
        STATE_ABANDONED = ConnRecord.State.ABANDONED
        with mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec:
            mock_conn_rec.query = mock.CoroutineMock()
            mock_conn_rec.Role = ConnRecord.Role
            mock_conn_rec.State = mock.MagicMock(
                COMPLETED=STATE_COMPLETED,
                INVITATION=STATE_INVITATION,
                ABANDONED=STATE_ABANDONED,
                get=mock.MagicMock(
                    side_effect=[
                        ConnRecord.State.ABANDONED,
                        ConnRecord.State.COMPLETED,
                        ConnRecord.State.INVITATION,
                    ]
                ),
            )
            conns = [  # in ascending order here
                mock.MagicMock(
                    serialize=mock.MagicMock(
                        return_value={
                            "state": ConnRecord.State.COMPLETED.rfc23,
                            "created_at": "1234567890",
                        }
                    )
                ),
                mock.MagicMock(
                    serialize=mock.MagicMock(
                        return_value={
                            "state": ConnRecord.State.INVITATION.rfc23,
                            "created_at": "1234567890",
                        }
                    )
                ),
                mock.MagicMock(
                    serialize=mock.MagicMock(
                        return_value={
                            "state": ConnRecord.State.ABANDONED.rfc23,
                            "created_at": "1234567890",
                        }
                    )
                ),
            ]
            mock_conn_rec.query.return_value = conns

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.connections_list(self.request)
                mock_conn_rec.query.assert_called_once_with(
                    ANY,
                    {
                        "invitation_id": "dummy",
                        "invitation_key": "some-invitation-key",
                        "their_public_did": "a_public_did",
                        "invitation_msg_id": "dummy_msg",
                    },
                    limit=100,
                    offset=0,
                    order_by="id",
                    descending=False,
                    post_filter_positive={
                        "their_role": list(ConnRecord.Role.REQUESTER.value),
                        "connection_protocol": "connections/1.0",
                    },
                    alt=True,
                )
                mock_response.assert_called_once_with(
                    {
                        "results": [
                            {
                                k: c.serialize.return_value[k]
                                for k in ["state", "created_at"]
                            }
                            for c in conns
                        ]
                    }  # sorted
                )

    async def test_connections_list_x(self):
        self.request.query = {
            "their_role": ConnRecord.Role.REQUESTER.rfc160,
            "alias": "my connection",
            "state": ConnRecord.State.COMPLETED.rfc23,
        }

        STATE_COMPLETED = ConnRecord.State.COMPLETED
        ROLE_REQUESTER = ConnRecord.Role.REQUESTER
        with mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec:
            mock_conn_rec.Role = mock.MagicMock(return_value=ROLE_REQUESTER)
            mock_conn_rec.State = mock.MagicMock(
                COMPLETED=STATE_COMPLETED,
                get=mock.MagicMock(return_value=ConnRecord.State.COMPLETED),
            )
            mock_conn_rec.query = mock.CoroutineMock(
                side_effect=test_module.StorageError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.connections_list(self.request)

    async def test_connections_retrieve(self):
        self.request.match_info = {"conn_id": "dummy"}
        mock_conn_rec = mock.MagicMock()
        mock_conn_rec.serialize = mock.MagicMock(return_value={"hello": "world"})

        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve_by_id,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_conn_rec_retrieve_by_id.return_value = mock_conn_rec

            await test_module.connections_retrieve(self.request)
            mock_response.assert_called_once_with({"hello": "world"})

    async def test_connections_endpoints(self):
        self.request.match_info = {"conn_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "BaseConnectionManager", autospec=True
            ) as mock_conn_mgr_cls,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_conn_mgr_cls.return_value = mock.MagicMock(
                get_endpoints=mock.CoroutineMock(
                    return_value=("localhost:8080", "1.2.3.4:8081")
                )
            )
            await test_module.connections_endpoints(self.request)
            mock_response.assert_called_once_with(
                {
                    "my_endpoint": "localhost:8080",
                    "their_endpoint": "1.2.3.4:8081",
                }
            )

    async def test_connections_endpoints_x(self):
        self.request.match_info = {"conn_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "BaseConnectionManager", autospec=True
            ) as mock_conn_mgr_cls,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_conn_mgr_cls.return_value = mock.MagicMock(
                get_endpoints=mock.CoroutineMock(side_effect=StorageNotFoundError())
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.connections_endpoints(self.request)

            mock_conn_mgr_cls.return_value = mock.MagicMock(
                get_endpoints=mock.CoroutineMock(side_effect=test_module.WalletError())
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.connections_endpoints(self.request)

    async def test_connections_metadata(self):
        self.request.match_info = {"conn_id": "dummy"}
        mock_conn_rec = mock.MagicMock()

        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve_by_id,
            mock.patch.object(
                mock_conn_rec, "metadata_get_all", mock.CoroutineMock()
            ) as mock_metadata_get_all,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_conn_rec_retrieve_by_id.return_value = mock_conn_rec
            mock_metadata_get_all.return_value = {"hello": "world"}

            await test_module.connections_metadata(self.request)
            mock_metadata_get_all.assert_called_once()
            mock_response.assert_called_once_with({"results": {"hello": "world"}})

    async def test_connections_metadata_get_single(self):
        self.request.match_info = {"conn_id": "dummy"}
        mock_conn_rec = mock.MagicMock()
        self.request.query = {"key": "test"}

        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve_by_id,
            mock.patch.object(mock_conn_rec, "metadata_get_all", mock.CoroutineMock()),
            mock.patch.object(
                mock_conn_rec, "metadata_get", mock.CoroutineMock()
            ) as mock_metadata_get,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_conn_rec_retrieve_by_id.return_value = mock_conn_rec
            mock_metadata_get.return_value = {"test": "value"}

            await test_module.connections_metadata(self.request)
            mock_metadata_get.assert_called_once()
            mock_response.assert_called_once_with({"results": {"test": "value"}})

    async def test_connections_metadata_x(self):
        self.request.match_info = {"conn_id": "dummy"}
        mock_conn_rec = mock.MagicMock()

        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve_by_id,
            mock.patch.object(
                mock_conn_rec, "metadata_get_all", mock.CoroutineMock()
            ) as mock_metadata_get_all,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_conn_rec_retrieve_by_id.return_value = mock_conn_rec
            mock_metadata_get_all.side_effect = StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.connections_metadata(self.request)

            mock_metadata_get_all.side_effect = test_module.BaseModelError()
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.connections_metadata(self.request)

    async def test_connections_metadata_set(self):
        self.request.match_info = {"conn_id": "dummy"}
        mock_conn_rec = mock.MagicMock()
        self.request.json = mock.CoroutineMock(
            return_value={"metadata": {"hello": "world"}}
        )

        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve_by_id,
            mock.patch.object(
                mock_conn_rec, "metadata_get_all", mock.CoroutineMock()
            ) as mock_metadata_get_all,
            mock.patch.object(
                mock_conn_rec, "metadata_set", mock.CoroutineMock()
            ) as mock_metadata_set,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_conn_rec_retrieve_by_id.return_value = mock_conn_rec
            mock_metadata_get_all.return_value = {"hello": "world"}

            await test_module.connections_metadata_set(self.request)
            mock_metadata_set.assert_called_once()
            mock_response.assert_called_once_with({"results": {"hello": "world"}})

    async def test_connections_metadata_set_x(self):
        self.request.match_info = {"conn_id": "dummy"}
        mock_conn_rec = mock.MagicMock()
        self.request.json = mock.CoroutineMock(
            return_value={"metadata": {"hello": "world"}}
        )

        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve_by_id,
            mock.patch.object(mock_conn_rec, "metadata_get_all", mock.CoroutineMock()),
            mock.patch.object(
                mock_conn_rec, "metadata_set", mock.CoroutineMock()
            ) as mock_metadata_set,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_conn_rec_retrieve_by_id.return_value = mock_conn_rec
            mock_metadata_set.side_effect = StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.connections_metadata_set(self.request)

            mock_metadata_set.side_effect = test_module.BaseModelError()
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.connections_metadata_set(self.request)

    async def test_connections_retrieve_not_found(self):
        self.request.match_info = {"conn_id": "dummy"}

        with mock.patch.object(
            test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve_by_id:
            mock_conn_rec_retrieve_by_id.side_effect = StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.connections_retrieve(self.request)

    async def test_connections_retrieve_x(self):
        self.request.match_info = {"conn_id": "dummy"}
        mock_conn_rec = mock.MagicMock()
        mock_conn_rec.serialize = mock.MagicMock(side_effect=test_module.BaseModelError())

        with mock.patch.object(
            test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve_by_id:
            mock_conn_rec_retrieve_by_id.return_value = mock_conn_rec

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.connections_retrieve(self.request)

    async def test_connections_remove(self):
        self.request.match_info = {"conn_id": "dummy"}
        mock_conn_rec = mock.MagicMock()
        mock_conn_rec.delete_record = mock.CoroutineMock()

        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve_by_id,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_conn_rec_retrieve_by_id.return_value = mock_conn_rec

            await test_module.connections_remove(self.request)
            mock_response.assert_called_once_with({})

    async def test_connections_remove_cache_key(self):
        cache = InMemoryCache()
        profile = self.context.profile
        await cache.set("conn_rec_state::dummy", "active")
        profile.context.injector.bind_instance(BaseCache, cache)
        self.request.match_info = {"conn_id": "dummy"}
        mock_conn_rec = mock.MagicMock()
        mock_conn_rec.delete_record = mock.CoroutineMock()
        assert (await cache.get("conn_rec_state::dummy")) == "active"
        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve_by_id,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_conn_rec_retrieve_by_id.return_value = mock_conn_rec

            await test_module.connections_remove(self.request)
            mock_response.assert_called_once_with({})
            assert not (await cache.get("conn_rec_state::dummy"))

    async def test_connections_remove_not_found(self):
        self.request.match_info = {"conn_id": "dummy"}

        with mock.patch.object(
            test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve_by_id:
            mock_conn_rec_retrieve_by_id.side_effect = StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.connections_remove(self.request)

    async def test_connections_remove_x(self):
        self.request.match_info = {"conn_id": "dummy"}
        mock_conn_rec = mock.MagicMock(
            delete_record=mock.CoroutineMock(side_effect=test_module.StorageError())
        )

        with mock.patch.object(
            test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve_by_id:
            mock_conn_rec_retrieve_by_id.return_value = mock_conn_rec

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.connections_remove(self.request)

    async def test_connections_create_static(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "my_seed": "my_seed",
                "my_did": "my_did",
                "their_seed": "their_seed",
                "their_did": "their_did",
                "their_verkey": "their_verkey",
                "their_endpoint": "their_endpoint",
                "their_role": "their_role",
                "alias": "alias",
            }
        )
        self.request.query = {
            "auto_accept": "true",
            "alias": "alias",
        }
        self.request.match_info = {"conn_id": "dummy"}

        mock_conn_rec = mock.MagicMock()
        mock_conn_rec.serialize = mock.MagicMock()
        mock_my_info = mock.MagicMock()
        mock_my_info.did = "my_did"
        mock_my_info.verkey = "my_verkey"
        mock_their_info = mock.MagicMock()
        mock_their_info.did = "their_did"
        mock_their_info.verkey = "their_verkey"

        with (
            mock.patch.object(
                test_module, "BaseConnectionManager", autospec=True
            ) as mock_conn_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_conn_mgr.return_value.create_static_connection = mock.CoroutineMock(
                return_value=(mock_my_info, mock_their_info, mock_conn_rec)
            )

            await test_module.connections_create_static(self.request)
            mock_response.assert_called_once_with(
                {
                    "my_did": mock_my_info.did,
                    "my_verkey": mock_my_info.verkey,
                    "their_did": mock_their_info.did,
                    "their_verkey": mock_their_info.verkey,
                    "my_endpoint": self.context.settings.get("default_endpoint"),
                    "record": mock_conn_rec.serialize.return_value,
                }
            )

    async def test_connections_create_static_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "my_seed": "my_seed",
                "my_did": "my_did",
                "their_seed": "their_seed",
                "their_did": "their_did",
                "their_verkey": "their_verkey",
                "their_endpoint": "their_endpoint",
                "their_role": "their_role",
                "alias": "alias",
            }
        )
        self.request.query = {
            "auto_accept": "true",
            "alias": "alias",
        }
        self.request.match_info = {"conn_id": "dummy"}

        mock_conn_rec = mock.MagicMock()
        mock_conn_rec.serialize = mock.MagicMock()
        mock_my_info = mock.MagicMock()
        mock_my_info.did = "my_did"
        mock_my_info.verkey = "my_verkey"
        mock_their_info = mock.MagicMock()
        mock_their_info.did = "their_did"
        mock_their_info.verkey = "their_verkey"

        with mock.patch.object(
            test_module, "BaseConnectionManager", autospec=True
        ) as mock_conn_mgr:
            mock_conn_mgr.return_value.create_static_connection = mock.CoroutineMock(
                side_effect=test_module.WalletError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.connections_create_static(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



"""RotateRecord model."""

from typing import Optional

from marshmallow import EXCLUDE, fields, validate

from .....messaging.models.base_record import BaseRecord, BaseRecordSchema


class RotateRecord(BaseRecord):
    """RotateRecord model.

    Stores state of a single DID rotate message exchange.
    """

    class Meta:
        """RotateRecord metadata."""

        schema_class = "RotateRecordSchema"

    RECORD_TYPE = "did_rotate"
    RECORD_ID_NAME = "record_id"

    # Role values
    ROLE_ROTATING = "rotating"
    ROLE_OBSERVING = "observing"

    # State values
    STATE_ROTATE_SENT = "rotate-sent"
    STATE_ROTATE_RECEIVED = "rotate-received"
    STATE_ACK_SENT = "ack-sent"
    STATE_ACK_RECEIVED = "ack-received"
    STATE_FAILED = "failed"

    TAG_NAMES = {"connection_id", "state", "role", "thread_id"}

    def __init__(
        self,
        *,
        record_id: Optional[str] = None,
        role: Optional[str] = None,
        state: Optional[str] = None,
        connection_id: Optional[str] = None,
        error: Optional[str] = None,
        new_did: Optional[str] = None,
        thread_id: Optional[str] = None,
        **kwargs,
    ):
        """Initialize a new RotateRecord."""
        super().__init__(record_id, state or self.STATE_ROTATE_SENT, **kwargs)
        self.role = role or self.ROLE_ROTATING
        self.connection_id = connection_id
        self.new_did = new_did
        self.thread_id = thread_id
        self.error = error

    @property
    def record_id(self) -> str:
        """Accessor for the ID associated with this record."""
        return self._id

    @classmethod
    async def retrieve_by_connection_id(
        cls, session, connection_id: str
    ) -> "RotateRecord":
        """Retrieve a rotate record by connection ID."""
        return await cls.retrieve_by_tag_filter(session, {"connection_id": connection_id})

    @classmethod
    async def retrieve_by_thread_id(cls, session, thread_id: str) -> "RotateRecord":
        """Retrieve a rotate record by thread ID."""
        return await cls.retrieve_by_tag_filter(session, {"thread_id": thread_id})

    @property
    def record_value(self) -> dict:
        """Accessor to for the JSON record value properties."""
        return {
            prop: getattr(self, prop)
            for prop in (
                "role",
                "state",
                "connection_id",
                "error",
                "new_did",
                "thread_id",
            )
        }


class RotateRecordSchema(BaseRecordSchema):
    """RotateRecord schema."""

    class Meta:
        """RotateRecordSchema metadata."""

        model_class = RotateRecord
        unknown = EXCLUDE

    record_id = fields.Str(required=True, metadata={"description": "Record identifier"})
    role = fields.Str(
        required=True,
        metadata={
            "description": "Role in the DID rotate protocol: rotating or observing"
        },
        validate=validate.OneOf(
            [RotateRecord.ROLE_ROTATING, RotateRecord.ROLE_OBSERVING]
        ),
    )
    state = fields.Str(
        required=True,
        metadata={
            "description": (
                "State of the DID rotate protocol: rotate-sent, rotate-received, "
                "ack-sent, ack-received, or failed"
            )
        },
        validate=validate.OneOf(
            [
                RotateRecord.STATE_ROTATE_SENT,
                RotateRecord.STATE_ROTATE_RECEIVED,
                RotateRecord.STATE_ACK_SENT,
                RotateRecord.STATE_ACK_RECEIVED,
                RotateRecord.STATE_FAILED,
            ]
        ),
    )
    connection_id = fields.Str(
        required=True,
        metadata={
            "description": "Connection identifier",
            "example": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        },
    )
    error = fields.Str(
        required=False,
        metadata={"description": "Error message", "example": "Invalid DID"},
    )
    new_did = fields.Str(
        required=False,
        metadata={
            "description": "New DID",
            "example": "did:sov:WRfXPg8dantKVubE3HX8pw",
        },
    )



"""Manager for multitenancy."""

import logging
from typing import Iterable, Optional

from ..askar.profile_anon import AskarAnonCredsProfile
from ..config.injection_context import InjectionContext
from ..config.wallet import wallet_config
from ..core.profile import Profile
from ..multitenant.base import BaseMultitenantManager
from ..wallet.models.wallet_record import WalletRecord
from .cache import ProfileCache

LOGGER = logging.getLogger(__name__)


class MultitenantManager(BaseMultitenantManager):
    """Class for handling multitenancy."""

    def __init__(self, profile: Profile):
        """Initialize default multitenant Manager.

        Args:
            profile: The profile for this manager
        """
        super().__init__(profile)
        self._profiles = ProfileCache(
            profile.settings.get_int("multitenant.cache_size") or 100
        )

    @property
    def open_profiles(self) -> Iterable[Profile]:
        """Return iterator over open profiles."""
        yield from self._profiles.profiles.values()

    async def get_wallet_profile(
        self,
        base_context: InjectionContext,
        wallet_record: WalletRecord,
        extra_settings: Optional[dict] = None,
        *,
        provision=False,
    ) -> Profile:
        """Get profile for a wallet record.

        Args:
            base_context: Base context to extend from
            wallet_record: Wallet record to get the context for
            extra_settings: Any extra context settings
            provision: Whether to provision the wallet

        Returns:
            Profile: Profile for the wallet record

        """
        extra_settings = extra_settings or {}
        wallet_id = wallet_record.wallet_id
        profile = self._profiles.get(wallet_id)
        if not profile:
            # Extend base context
            context = base_context.copy()

            # Settings we don't want to use from base wallet
            reset_settings = {
                "wallet.recreate": False,
                "wallet.seed": None,
                "wallet.rekey": None,
                "wallet.name": None,
                "wallet.type": None,
                "mediation.open": None,
                "mediation.invite": None,
                "mediation.default_id": None,
                "mediation.clear": None,
            }
            extra_settings["admin.webhook_urls"] = self.get_webhook_urls(
                base_context, wallet_record
            )

            context.settings = (
                context.settings.extend(reset_settings)
                .extend(wallet_record.settings)
                .extend(extra_settings)
            )

            # MTODO: add ledger config
            profile, _ = await wallet_config(context, provision=provision)
            self._profiles.put(wallet_id, profile)

        # return anoncreds profile if explicitly set as wallet type
        if profile.context.settings.get("wallet.type") == "askar-anoncreds":
            return AskarAnonCredsProfile(
                profile.opened,
                profile.context,
            )

        return profile

    async def update_wallet(self, wallet_id: str, new_settings: dict) -> WalletRecord:
        """Update an existing wallet and wallet record.

        Args:
            wallet_id: The wallet id of the wallet record
            new_settings: The context settings to be updated for this wallet

        Returns:
            WalletRecord: The updated wallet record

        """
        wallet_record = await super().update_wallet(wallet_id, new_settings)

        # Wallet record has been updated but profile settings in memory must
        # also be refreshed; update profile only if loaded
        profile = self._profiles.get(wallet_id)
        if profile:
            profile.settings.update(wallet_record.settings)

            extra_settings = {
                "admin.webhook_urls": self.get_webhook_urls(
                    self._profile.context, wallet_record
                ),
            }
            profile.settings.update(extra_settings)

        return wallet_record

    async def remove_wallet_profile(self, profile: Profile):
        """Remove the wallet profile instance.

        Args:
            profile: The wallet profile instance

        """
        wallet_id = profile.settings.get_str("wallet.id")
        self._profiles.remove(wallet_id)
        await profile.remove()



from tempfile import NamedTemporaryFile
from unittest import IsolatedAsyncioTestCase, mock
from weakref import ref

from ...utils.stats import Collector
from ..base import InjectionError
from ..injection_context import InjectionContext
from ..provider import (
    BaseProvider,
    CachedProvider,
    ClassProvider,
    InstanceProvider,
    StatsProvider,
)
from ..settings import Settings


class X:
    def __init__(self, *args, **kwargs):
        pass


class TestProvider(IsolatedAsyncioTestCase):
    async def test_stats_provider_init_x(self):
        """Cover stats provider init error on no provider."""
        with self.assertRaises(ValueError):
            StatsProvider(None, ["method"])

    async def test_stats_provider_provide_collector(self):
        """Cover call to provide with collector."""

        timing_log = NamedTemporaryFile().name
        settings = {"timing.enabled": True, "timing.log.file": timing_log}
        mock_provider = mock.MagicMock(BaseProvider, autospec=True)
        mock_provider.provide.return_value.mock_method = lambda: ()
        stats_provider = StatsProvider(mock_provider, ("mock_method",))
        collector = Collector(log_path=timing_log)

        context = InjectionContext(settings=settings, enforce_typing=False)
        context.injector.bind_instance(Collector, collector)

        stats_provider.provide(Settings(settings), context.injector)

    async def test_cached_provider_same_unique_settings(self):
        """Cover same unique keys returns same instance."""
        first_settings = Settings(
            {"wallet.name": "wallet.name", "wallet.key": "wallet.key"}
        )
        second_settings = first_settings.extend({"wallet.key": "another.wallet.key"})

        cached_provider = CachedProvider(
            ClassProvider("acapy_agent.config.settings.Settings"), ("wallet.name",)
        )
        context = InjectionContext()

        first_instance = cached_provider.provide(first_settings, context.injector)
        second_instance = cached_provider.provide(second_settings, context.injector)

        assert first_instance is second_instance

    async def test_cached_provider_different_unique_settings(self):
        """Cover two different unique keys returns different instance."""
        first_settings = Settings(
            {"wallet.name": "wallet.name", "wallet.key": "wallet.key"}
        )
        second_settings = first_settings.extend({"wallet.name": "another.wallet.name"})

        cached_provider = CachedProvider(
            ClassProvider("acapy_agent.config.settings.Settings"), ("wallet.name",)
        )
        context = InjectionContext()

        first_instance = cached_provider.provide(first_settings, context.injector)
        second_instance = cached_provider.provide(second_settings, context.injector)

        assert first_instance is not second_instance

    async def test_instance_provider_ref_gone_x(self):
        context = InjectionContext(settings=Settings({}), enforce_typing=False)
        item = X()
        instance = ref(item)
        del item

        with self.assertRaises(InjectionError):
            provider = InstanceProvider(instance=instance)
            provider.provide(Settings({}), context.injector)

    async def test_class_provider_ref_gone_x(self):
        context = InjectionContext(settings=Settings({}), enforce_typing=False)
        item = X()
        instance = ref(item)
        del item

        with self.assertRaises(InjectionError):
            provider = ClassProvider(X, instance)
            provider.provide(Settings({}), context.injector)

        with self.assertRaises(InjectionError):
            provider = ClassProvider(X, kwarg=instance)
            provider.provide(Settings({}), context.injector)



CREDENTIALS_V1 = {
    "@context": {
        "@version": 1.1,
        "@protected": True,
        "id": "@id",
        "type": "@type",
        "VerifiableCredential": {
            "@id": "https://www.w3.org/2018/credentials#VerifiableCredential",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "cred": "https://www.w3.org/2018/credentials#",
                "sec": "https://w3id.org/security#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                "credentialSchema": {
                    "@id": "cred:credentialSchema",
                    "@type": "@id",
                    "@context": {
                        "@version": 1.1,
                        "@protected": True,
                        "id": "@id",
                        "type": "@type",
                        "cred": "https://www.w3.org/2018/credentials#",
                        "JsonSchemaValidator2018": "cred:JsonSchemaValidator2018",
                    },
                },
                "credentialStatus": {"@id": "cred:credentialStatus", "@type": "@id"},
                "credentialSubject": {"@id": "cred:credentialSubject", "@type": "@id"},
                "evidence": {"@id": "cred:evidence", "@type": "@id"},
                "expirationDate": {
                    "@id": "cred:expirationDate",
                    "@type": "xsd:dateTime",
                },
                "holder": {"@id": "cred:holder", "@type": "@id"},
                "issued": {"@id": "cred:issued", "@type": "xsd:dateTime"},
                "issuer": {"@id": "cred:issuer", "@type": "@id"},
                "issuanceDate": {"@id": "cred:issuanceDate", "@type": "xsd:dateTime"},
                "proof": {"@id": "sec:proof", "@type": "@id", "@container": "@graph"},
                "refreshService": {
                    "@id": "cred:refreshService",
                    "@type": "@id",
                    "@context": {
                        "@version": 1.1,
                        "@protected": True,
                        "id": "@id",
                        "type": "@type",
                        "cred": "https://www.w3.org/2018/credentials#",
                        "ManualRefreshService2018": "cred:ManualRefreshService2018",
                    },
                },
                "termsOfUse": {"@id": "cred:termsOfUse", "@type": "@id"},
                "validFrom": {"@id": "cred:validFrom", "@type": "xsd:dateTime"},
                "validUntil": {"@id": "cred:validUntil", "@type": "xsd:dateTime"},
            },
        },
        "VerifiablePresentation": {
            "@id": "https://www.w3.org/2018/credentials#VerifiablePresentation",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "cred": "https://www.w3.org/2018/credentials#",
                "sec": "https://w3id.org/security#",
                "holder": {"@id": "cred:holder", "@type": "@id"},
                "proof": {"@id": "sec:proof", "@type": "@id", "@container": "@graph"},
                "verifiableCredential": {
                    "@id": "cred:verifiableCredential",
                    "@type": "@id",
                    "@container": "@graph",
                },
            },
        },
        "EcdsaSecp256k1Signature2019": {
            "@id": "https://w3id.org/security#EcdsaSecp256k1Signature2019",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "sec": "https://w3id.org/security#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                "challenge": "sec:challenge",
                "created": {
                    "@id": "http://purl.org/dc/terms/created",
                    "@type": "xsd:dateTime",
                },
                "domain": "sec:domain",
                "expires": {"@id": "sec:expiration", "@type": "xsd:dateTime"},
                "jws": "sec:jws",
                "nonce": "sec:nonce",
                "proofPurpose": {
                    "@id": "sec:proofPurpose",
                    "@type": "@vocab",
                    "@context": {
                        "@version": 1.1,
                        "@protected": True,
                        "id": "@id",
                        "type": "@type",
                        "sec": "https://w3id.org/security#",
                        "assertionMethod": {
                            "@id": "sec:assertionMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "authentication": {
                            "@id": "sec:authenticationMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                    },
                },
                "proofValue": "sec:proofValue",
                "verificationMethod": {"@id": "sec:verificationMethod", "@type": "@id"},
            },
        },
        "EcdsaSecp256r1Signature2019": {
            "@id": "https://w3id.org/security#EcdsaSecp256r1Signature2019",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "sec": "https://w3id.org/security#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                "challenge": "sec:challenge",
                "created": {
                    "@id": "http://purl.org/dc/terms/created",
                    "@type": "xsd:dateTime",
                },
                "domain": "sec:domain",
                "expires": {"@id": "sec:expiration", "@type": "xsd:dateTime"},
                "jws": "sec:jws",
                "nonce": "sec:nonce",
                "proofPurpose": {
                    "@id": "sec:proofPurpose",
                    "@type": "@vocab",
                    "@context": {
                        "@version": 1.1,
                        "@protected": True,
                        "id": "@id",
                        "type": "@type",
                        "sec": "https://w3id.org/security#",
                        "assertionMethod": {
                            "@id": "sec:assertionMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "authentication": {
                            "@id": "sec:authenticationMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                    },
                },
                "proofValue": "sec:proofValue",
                "verificationMethod": {"@id": "sec:verificationMethod", "@type": "@id"},
            },
        },
        "Ed25519Signature2018": {
            "@id": "https://w3id.org/security#Ed25519Signature2018",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "sec": "https://w3id.org/security#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                "challenge": "sec:challenge",
                "created": {
                    "@id": "http://purl.org/dc/terms/created",
                    "@type": "xsd:dateTime",
                },
                "domain": "sec:domain",
                "expires": {"@id": "sec:expiration", "@type": "xsd:dateTime"},
                "jws": "sec:jws",
                "nonce": "sec:nonce",
                "proofPurpose": {
                    "@id": "sec:proofPurpose",
                    "@type": "@vocab",
                    "@context": {
                        "@version": 1.1,
                        "@protected": True,
                        "id": "@id",
                        "type": "@type",
                        "sec": "https://w3id.org/security#",
                        "assertionMethod": {
                            "@id": "sec:assertionMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "authentication": {
                            "@id": "sec:authenticationMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                    },
                },
                "proofValue": "sec:proofValue",
                "verificationMethod": {"@id": "sec:verificationMethod", "@type": "@id"},
            },
        },
        "RsaSignature2018": {
            "@id": "https://w3id.org/security#RsaSignature2018",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "challenge": "sec:challenge",
                "created": {
                    "@id": "http://purl.org/dc/terms/created",
                    "@type": "xsd:dateTime",
                },
                "domain": "sec:domain",
                "expires": {"@id": "sec:expiration", "@type": "xsd:dateTime"},
                "jws": "sec:jws",
                "nonce": "sec:nonce",
                "proofPurpose": {
                    "@id": "sec:proofPurpose",
                    "@type": "@vocab",
                    "@context": {
                        "@version": 1.1,
                        "@protected": True,
                        "id": "@id",
                        "type": "@type",
                        "sec": "https://w3id.org/security#",
                        "assertionMethod": {
                            "@id": "sec:assertionMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "authentication": {
                            "@id": "sec:authenticationMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                    },
                },
                "proofValue": "sec:proofValue",
                "verificationMethod": {"@id": "sec:verificationMethod", "@type": "@id"},
            },
        },
        "proof": {
            "@id": "https://w3id.org/security#proof",
            "@type": "@id",
            "@container": "@graph",
        },
    }
}



from unittest import TestCase

from ..pres_request_schema import DIFProofRequestSchema


class TestPresRequestSchema(TestCase):
    """DIF Presentation Request Test"""

    def test_limit_disclosure(self):
        test_pd_a = {
            "options": {
                "challenge": "3fa85f64-5717-4562-b3fc-2c963f66afa7",
                "domain": "4jt78h47fh47",
            },
            "presentation_definition": {
                "id": "32f54163-7166-48f1-93d8-ff217bdb0654",
                "submission_requirements": [
                    {
                        "name": "Citizenship Information",
                        "rule": "pick",
                        "min": 1,
                        "from": "A",
                    }
                ],
                "input_descriptors": [
                    {
                        "id": "citizenship_input_1",
                        "name": "EU Driver's License",
                        "group": ["A"],
                        "schema": [
                            {
                                "uri": "https://www.w3.org/2018/credentials#VerifiableCredential"
                            }
                        ],
                        "constraints": {
                            "limit_disclosure": "required",
                            "fields": [
                                {
                                    "path": ["$.credentialSubject.givenName"],
                                    "purpose": "The claim must be from one of the specified issuers",
                                    "filter": {
                                        "type": "string",
                                        "enum": ["JOHN", "CAI"],
                                    },
                                }
                            ],
                        },
                    }
                ],
            },
        }
        test_pd_b = {
            "options": {
                "challenge": "3fa85f64-5717-4562-b3fc-2c963f66afa7",
                "domain": "4jt78h47fh47",
            },
            "presentation_definition": {
                "id": "32f54163-7166-48f1-93d8-ff217bdb0654",
                "submission_requirements": [
                    {
                        "name": "Citizenship Information",
                        "rule": "pick",
                        "min": 1,
                        "from": "A",
                    }
                ],
                "input_descriptors": [
                    {
                        "id": "citizenship_input_1",
                        "name": "EU Driver's License",
                        "group": ["A"],
                        "schema": [
                            {
                                "uri": "https://www.w3.org/2018/credentials#VerifiableCredential"
                            }
                        ],
                        "constraints": {
                            "limit_disclosure": "preferred",
                            "fields": [
                                {
                                    "path": ["$.credentialSubject.givenName"],
                                    "purpose": "The claim must be from one of the specified issuers",
                                    "filter": {
                                        "type": "string",
                                        "enum": ["JOHN", "CAI"],
                                    },
                                }
                            ],
                        },
                    }
                ],
            },
        }

        pres_request_a = DIFProofRequestSchema().load(test_pd_a)
        test_limit_disclosure_a = (
            pres_request_a.presentation_definition.input_descriptors[
                0
            ].constraint.limit_disclosure
        )
        assert test_limit_disclosure_a == "required"
        pres_request_b = DIFProofRequestSchema().load(test_pd_b)
        test_limit_disclosure_b = (
            pres_request_b.presentation_definition.input_descriptors[
                0
            ].constraint.limit_disclosure
        )
        assert test_limit_disclosure_b == "preferred"
